//
//	File: PhotoshopShell.c
//
//	portions Copyright (c) 1990-1, Thomas Knoll.
//	portions Copyright (c) 1992-6, Adobe Systems Incorporated.
//	portions Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
//	All rights reserved
//
//	This file incorporates some portions of versions 3.0.4, 4.0, 
//	a few lines of 5.0, and the plug-in modes of 9.0 from Photoshop's Filter 
//	module interface. It also incorporates the API described in Russ&Russ, 2007
//
//	This code may be used by purchasers of the book for educational and personal research
//	purposes but not for any commercial purpose.  The Copyright notices must be preserved
//	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
//  distributed for a fee.
//	


#include "PhotoshopShell.h"
#include "PiPLdata.h"
#include <math.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <io.h>
#include <stdio.h> 

// PRIVATE STRUCTURES
typedef struct Globals
{
	long 		 mode;				//from photoshop's list of modes
	OSErr 		 result;			//for error propagation
	Edit_PSData *stuff;				//a photoshop filter calling record
	long		 width;				//image width (pixels)
	long		 height;			//image height (pixels)
	long		 bitdepth;			//either 8, 16, or 32 bits per channel
	long		 planeCount;		//number of planes (channels) we're requesting from Photoshop, 
									//depends upon mode
	Ptr			 cachedWritePtr;	//line waiting to be written during next AdvanceState
	long		 cachedWriteLineNum;//line number for cached write
	long		 cachedLineLength;	//bytes
	Boolean		 firstTimeAdvanceState;	//first time through AdvanceState
	Boolean		 hasCachedData;		//whether there is data waiting to be written
} Globals, *GPtr;

typedef struct _FSID //OPAQUE type File_ID == int32. Implementation is platform specific
{
	HANDLE		 fref;		//actual file reference (hidden from user)
	Boolean		 read;		//true = read, false = write
	Boolean		 open;		//true = still open, false = closed
	struct _FSID *nextFile;	//linked list of open files
} _FSID, *FSIDptr;

Globals gPrimaryImage;			//PRIVATE GLOBALS
_FSID	*gOpenFileList = NULL;	//Linked list of currently open text files

typedef struct __PointerPad
{	//16 bytes
	OSType			signature1;	//'TaiL'
	long			length;
	long			size;
	struct __PointerPad   *head;
} __PointerPad;

__PointerPad *gPointerBlockListHead = NULL;
__PointerPad *gCurrentBlock = NULL;
long		  gBlockPointerCount = 0;

// LOCAL/PRIVATE PROTOTYPES
OSErr	_PhotoshopAdvanceState();				//noErr = good
void	_DoProgress(int32 value, int32 total);
Boolean	_CommandPeriod();
void	_CloseAllTextFiles();
void	_RRStartupInitialization();
void	_RRShutdownCleanup();

//R&R Glue code

/////////////////////////////////////////////////////////////////////////////////
//
//	CreateAPointer
//
////////////////////////////////////////////////////////////////////////////////
void *CreateAPointer(long length, long size)
{	//additional pointer error protection could be added here
	if ((uint32)length * (uint32)size == 0)
	{
		RR_ERRORSTRING("Attempt to allocate a zero-length pointer. (CreateAPointer)");
	}
	return malloc((uint32)length * (uint32)size);
}//CreateAPointer

/////////////////////////////////////////////////////////////////////////////////
//
//	CreateAPointerClear
//
////////////////////////////////////////////////////////////////////////////////
void *CreateAPointerClear(long length, long size)
{
	void *ap = CreateAPointer(length, size);
	uint32	i;
	if (!ap)	return NULL;

	switch (size) //clear everything out
	{
		case sizeof(double): //FFTelem ~ double = 8
		{
			double *dummyd = (double *)ap;
			for (i = 0; i < (uint32)length; i++)
			{
				*dummyd = 0;
				dummyd++;
			}
		}
		break;

		case sizeof(uint32): //4
		case (3 * sizeof(uint32)): //12
		{
			uint32 *dummyl = (uint32 *)ap;
			for (i = 0; i < (uint32)length * (uint32)size / sizeof(uint32); i++)
			{
				*dummyl = 0;
				dummyl++;
			}
		}
		break;

		case (3 * sizeof(uint16)): //6
		case sizeof(uint16): //2
		{
			uint16 *dummys = (uint16 *)ap;
			for (i = 0; i < (uint32)length * (uint32)size / sizeof(uint16); i++)
			{
				*dummys = 0;
				dummys++;
			}
		}
		break;

		case 1:
		default:
		{
			char *dummyc = (char *)ap;
			for (i = 0; i < (uint32)length * (uint32)size; i++)
				dummyc[i] = 0;
		}
		break;
	}//switch

	return ap;
}//CreateAPointerClear


/////////////////////////////////////////////////////////////////////////////////
//
//	DisposeAPointer
//
////////////////////////////////////////////////////////////////////////////////
void	DisposeAPointer(void *ptr)
{	//additional pointer error protection could be added here
	if (!ptr)
	{
		RR_ERRORSTRING("Attempt to delete NULL pointer");
		return;
	}
	free(ptr);
}//DisposeAPointer

////////////////////////////////////////////////////////////////////////////////
//
//	ErrorMessage
//
////////////////////////////////////////////////////////////////////////////////
void	ErrorMessage(char *bstring)
{
	(void)MessageBox((HWND)NULL, (LPSTR)bstring, (LPSTR)"                                         ", 
					  MB_OK | MB_ICONHAND | MB_ICONEXCLAMATION);
	return;
}//ErrorMessage

////////////////////////////////////////////////////////////////////////////////
//
//	GetOriginalDimensions
//
////////////////////////////////////////////////////////////////////////////////
void GetOriginalDimensions(long *width, long *height)
{
	*width = gPrimaryImage.width;
	*height = gPrimaryImage.height;
}//GetOriginalDimensions


////////////////////////////////////////////////////////////////////////////////
//
//	Private function: _RGB96toLAB96
//
////////////////////////////////////////////////////////////////////////////////
void 	_RGB96toLAB96(RGBPixel *data, long width);		//conversion in place, (not very well)
void 	_RGB96toLAB96(RGBPixel *data, long width)
{
	long	x;
	for (x = 0; x < width; x++)
	{
		float R = data[x].red;
		float G = data[x].green;
		float B = data[x].blue;
		float L = (R + G + B + (float)1.0) / (float)3.0;	//has to stay in range
		float a = (float).33333*(R - G) + (float)127.5;			
		float b = (float).83333*((R + G) / (float)2.0 - B) + (float)127.5;
		//pseudo adobe perceptual correction
		L = (float)3.032 * ((float)exp(log(L+0.05)*(float)0.8)-(float)0.09102821);	//gamma 0.8
		if (a < 0) a = 0;		else if (a > 255) a = 255;
		if (b < 0) b = 0;		else if (b > 255) b = 255;
		data[x].red =	L;
		data[x].green = a;
		data[x].blue =  b;
	}
}//_RGB96toLAB96

////////////////////////////////////////////////////////////////////////////////
//
//	Private function: _LAB96toRGB96
//
////////////////////////////////////////////////////////////////////////////////
void 	_LAB96toRGB96(RGBPixel *data, long width);		//conversion in place, (not very well)
void 	_LAB96toRGB96(RGBPixel *data, long width)
{
	long	x;
	for (x = 0; x < width; x++)
	{
		float L = data[x].red;
		float a = (data[x].green - (float)127.5)*(float)3;
		float b = (data[x].blue - (float)127.5)*(float)1.2;
		float R, G, B;
		//pseudo adobe perceptual correction
		L = ((float)exp(log(L+(float)0.05)*(float)1.28)-(float)0.216113355)/(float)4.71913337521; //gamma 1.28
		R = (L + a/(float)2.0 + b/(float)3.0 + (float)0.5);// L*a*b model
		G = (L - a/(float)2.0 + b/(float)3.0 + (float)0.5);
		B = (L - (float)2.0 * b/(float)3.0 + (float)0.5);
		if (R < 0) R = 0;		else if (R > 255) R = 255;
		if (G < 0) G = 0;		else if (G > 255) G = 255;
		if (B < 0) B = 0;		else if (B > 255) B = 255;
		data[x].red =   R;
		data[x].green = G;
		data[x].blue =  B;
	}
}//_LAB96toRGB96


////////////////////////////////////////////////////////////////////////////////
//
//	ReadOriginalLine
//
////////////////////////////////////////////////////////////////////////////////
void	ReadOriginalLine(long yline, RGBPixel *line)
{
	long	x;
	long	width = gPrimaryImage.width;
	unsigned char *iPtr;
	unsigned short *sPtr;
	float *fPtr;

	//deal with first time through and setting up cached writes
	if (gPrimaryImage.firstTimeAdvanceState)
	{
		//make initial line 0 transaction take place -- was set to read line #0, outbuffer set up to write line #0 next time
		if (_PhotoshopAdvanceState() != noErr)
		{
			RR_ERRORSTRING("AdvanceState function is not available.  Plug-in cannot run. (ReadOriginalLine)");
			return;
		}
		gPrimaryImage.cachedLineLength = gPrimaryImage.stuff->inRowBytes;									//get length of data in buffer
		memcpy(gPrimaryImage.cachedWritePtr, gPrimaryImage.stuff->inData, gPrimaryImage.stuff->inRowBytes);	//clone line #0 into outData
		memcpy(gPrimaryImage.stuff->outData, gPrimaryImage.cachedWritePtr, gPrimaryImage.stuff->inRowBytes);
		gPrimaryImage.cachedWriteLineNum = 0;																//we've queued up line #0
		gPrimaryImage.firstTimeAdvanceState = FALSE;
		gPrimaryImage.hasCachedData = FALSE;	//line#0 is now in the outData buffer, waiting to be written
	}//firstTimeAdvanceState

	//bounds test
	if ((yline < 0) || (yline > gPrimaryImage.height - 1))
	{
		for (x = 0; x < gPrimaryImage.width; x++)
		{
			line[x].red = line[x].green = line[x].blue = 0.0;
		}
		RR_ERRORSTRING("Attempt to read beyond bounds of original image. (ReadOriginalLine)");
		return;
	}
	//Specific appropriate rectangle for input (inData)
	gPrimaryImage.stuff->inRect.top = (short)yline;
	gPrimaryImage.stuff->inRect.bottom = (short)yline + 1;

	if (gPrimaryImage.hasCachedData) //request new outData buffer for next time (do not overwrite if there is stuff NOW)
	{
		gPrimaryImage.stuff->outRect.top = (short)gPrimaryImage.cachedWriteLineNum;
		gPrimaryImage.stuff->outRect.bottom = (short)gPrimaryImage.cachedWriteLineNum + 1;
	}

	if (_PhotoshopAdvanceState() != noErr)	//make Read/Write transaction take place with Photoshop
	{
		RR_ERRORSTRING("AdvanceState function is not available.  Plug-in cannot run. (ReadOriginalLine)");
		return;
	}
	//Point to Photoshop's buffer so we can parse what was just read 
	iPtr = (unsigned char *)gPrimaryImage.stuff->inData;
	sPtr = (unsigned short *)iPtr;
	fPtr = (float *)iPtr;
	//convert data into RGB96 within user's line buffer
	switch (gPrimaryImage.mode)
	{
		case plugInModeGrayScale: 		//supported
			for (x = 0; x < width; x++)
			{
				line[x].red = line[x].green = line[x].blue = (float)iPtr[x];
			}
		break;

		case plugInModeLabColor:		//supported via transform
		case plugInModeRGBColor:		//supported
			for (x = 0; x < width; x++)
			{
				line[x].red =	(float)iPtr[x * 3 + 0];
				line[x].green = (float)iPtr[x * 3 + 1];
				line[x].blue =  (float)iPtr[x * 3 + 2];
			}
			if (gPrimaryImage.mode == plugInModeLab48)
				_LAB96toRGB96(line, width);
		break;

		case plugInModeGray16:			//supported
			for (x = 0; x < width; x++)
			{
				line[x].red = line[x].green = line[x].blue = ((float)sPtr[x]) * (float)(255.0 / 32768.0);
			}
		break;

		case plugInModeLab48:			//supported via transform
		case plugInModeRGB48:			//supported
			for (x = 0; x < width; x++)
			{
				line[x].red =	((float)sPtr[x * 3 + 0]) * (float)(255.0 / 32768.0);
				line[x].green = ((float)sPtr[x * 3 + 1]) * (float)(255.0 / 32768.0);
				line[x].blue =  ((float)sPtr[x * 3 + 2]) * (float)(255.0 / 32768.0);
			}
			if (gPrimaryImage.mode == plugInModeLab48)
				_LAB96toRGB96(line, width);
		break;

		case plugInModeRGB96:			//supported
			for (x = 0; x < width; x++)
			{
				line[x].red =	fPtr[x * 3 + 0] * (float)(255.0 / 1.0);
				line[x].green = fPtr[x * 3 + 1] * (float)(255.0 / 1.0);
				line[x].blue =  fPtr[x * 3 + 2] * (float)(255.0 / 1.0);
			}
		break;

		case plugInModeGray32:      	//supported
			for (x = 0; x < width; x++)
			{
				line[x].red = line[x].green = line[x].blue = fPtr[x] * (float)(255.0 / 1.0);
			}
		break;

		case plugInModeCMYKColor:		case plugInModeCMYK64:		case plugInModeDeepMultichannel:
		case plugInModeMultichannel:	case plugInModeDuotone16:	case plugInModeDuotone:	 
		case plugInModeHSLColor:		case plugInModeHSBColor:	case plugInModeIndexedColor:
		case plugInModeBitmap:			default:
			RR_ERRORSTRING("Unsupported image mode.  Plug-in cannot run. (ReadOriginalLine)");
		break;	
	}//switch (mode)

	//setup for next cached write
	if (gPrimaryImage.hasCachedData) //there is cached data to be written to outData
	{
		memcpy(gPrimaryImage.stuff->outData, gPrimaryImage.cachedWritePtr,  gPrimaryImage.cachedLineLength);
		gPrimaryImage.hasCachedData = FALSE;	//just copied from my buffer into the outData
	}
}//ReadOriginalLine


////////////////////////////////////////////////////////////////////////////////
//
//	WriteResultLine
//
////////////////////////////////////////////////////////////////////////////////
void	WriteResultLine(long yline, RGBPixel *line)
{
	long	x;
	long	width = gPrimaryImage.width;
	unsigned char *iPtr;
	unsigned short *sPtr;
	float *fPtr;
	
	//deal with first time through and setting up cached writes
	if (gPrimaryImage.firstTimeAdvanceState)
	{
		//make initial line 0 transaction take place -- set to read line #0, outbuffer set up to write line #0 next time
		if (_PhotoshopAdvanceState() != noErr)
		{
			RR_ERRORSTRING("AdvanceState function is not available.  firstTime = TRUE. (WriteResultLine)");
			return;
		}
		gPrimaryImage.cachedLineLength = gPrimaryImage.stuff->inRowBytes;
		memcpy(gPrimaryImage.cachedWritePtr, gPrimaryImage.stuff->inData, gPrimaryImage.stuff->inRowBytes);
		memcpy(gPrimaryImage.stuff->outData, gPrimaryImage.cachedWritePtr, gPrimaryImage.stuff->inRowBytes);
		gPrimaryImage.cachedWriteLineNum = 0;																//we've queued up line #0
		gPrimaryImage.firstTimeAdvanceState = FALSE;
		gPrimaryImage.hasCachedData = FALSE;	//line#0 is now in the outData buffer, waiting to be written
	}//firstTimeAdvanceState

	//bounds test
	if ((yline < 0) || (yline > gPrimaryImage.height - 1))
	{
		RR_ERRORSTRING("Attempt to write beyond bounds of original image. (WriteResultLine)");
		return;
	}

	//if there is a cached write, then flush with a dummy read on the last read line
	if (gPrimaryImage.hasCachedData)
	{
		//set up for the cached buffer request (which will flush current contents
		gPrimaryImage.stuff->outRect.top = (short)gPrimaryImage.cachedWriteLineNum;									
		gPrimaryImage.stuff->outRect.bottom = (short)gPrimaryImage.cachedWriteLineNum + 1;
		if (_PhotoshopAdvanceState() != noErr)
		{
			RR_ERRORSTRING("AdvanceState function is not available.  Plug-in cannot run. (WriteResultLine)");
			return;
		}
		//now fill the outData buffer with the cache for THIS time
		memcpy(gPrimaryImage.stuff->outData, gPrimaryImage.cachedWritePtr,  gPrimaryImage.cachedLineLength);
		gPrimaryImage.hasCachedData = FALSE;	//just copied from my buffer into the outData
	}

	//now, request yline to be written next time
	gPrimaryImage.stuff->outRect.top = (short)yline;									
	gPrimaryImage.stuff->outRect.bottom = (short)yline + 1;
	gPrimaryImage.cachedWriteLineNum = yline;
	gPrimaryImage.hasCachedData = TRUE;	//we're about to have cached data again

	//Point to OUR cache buffer
	iPtr = (unsigned char *)gPrimaryImage.cachedWritePtr;
	sPtr = (unsigned short *)iPtr;
	fPtr = (float *)iPtr;

	//convert from RGB96 in user's line to our buffer
	switch (gPrimaryImage.mode)
	{
		case plugInModeGrayScale: 		//supported
			for (x = 0; x < width; x++)
			{
				float value = (line[x].red + line[x].green + line[x].blue) / (float)3.0;
				if (value < 0) value = 0;
				if (value > 255) value = 255;
				iPtr[x] = (unsigned char)value;
			}
		break;

		case plugInModeLabColor:		//supported via transform
		case plugInModeRGBColor:		//supported
			if (gPrimaryImage.mode == plugInModeLabColor)
				_RGB96toLAB96(line, width);
			for (x = 0; x < width; x++)
			{
				float value = line[x].red;
				if (value < 0) value = 0;
				if (value > 255) value = 255;
				iPtr[x*3 + 0] = (unsigned char)value;
				value = line[x].green;
				if (value < 0) value = 0;
				if (value > 255) value = 255;
				iPtr[x*3 + 1] = (unsigned char)value;
				value = line[x].blue;
				if (value < 0) value = 0;
				if (value > 255) value = 255;
				iPtr[x*3 + 2] = (unsigned char)value;
			}
		break;

		case plugInModeGray16:			//supported
			for (x = 0; x < width; x++)
			{
				float value = (float)(32768.0 / 255.0) * (line[x].red + line[x].green + line[x].blue) / (float)3.0;
				if (value < 0) value = 0;
				if (value > 32768) value = 32768;
				sPtr[x] = (unsigned short)value;
			}
		break;

		case plugInModeLab48:			//supported via transform
		case plugInModeRGB48:			//supported
			if (gPrimaryImage.mode == plugInModeLabColor)
				_RGB96toLAB96(line, width);
			for (x = 0; x < width; x++)
			{
				float value = line[x].red * (float)(32768.0 / 255.0);
				if (value < 0) value = 0;
				if (value > 32768) value = 32768;
				sPtr[x*3 + 0] = (unsigned short)value;
				value = line[x].green * (float)(32768.0 / 255.0);
				if (value < 0) value = 0;
				if (value > 32768) value = 32768;
				sPtr[x*3 + 1] = (unsigned short)value;
				value = line[x].blue * (float)(32768.0 / 255.0);
				if (value < 0) value = 0;
				if (value > 32768) value = 32768;
				sPtr[x*3 + 2] = (unsigned short)value;
			}
		break;

		case plugInModeRGB96:			//supported
			for (x = 0; x < width; x++)
			{	//no bounds checking in 32-bit mode, they use 0..1 as the "normal" range
				//but, Photoshop allows values from -inf to +inf for HDR images
				fPtr[x * 3 + 0] = line[x].red / (float)255.0;
				fPtr[x * 3 + 1] = line[x].green / (float)255.0;
				fPtr[x * 3 + 2] = line[x].blue / (float)255.0;
			}
		break;

		case plugInModeGray32:      	//supported
			for (x = 0; x < width; x++)
			{
				fPtr[x] = (line[x].red + line[x].green + line[x].blue) / (float)(255.0 * 3.0);
			}
		break;

		case plugInModeCMYKColor:		case plugInModeCMYK64:		case plugInModeDeepMultichannel:
		case plugInModeMultichannel:	case plugInModeDuotone16:	case plugInModeDuotone:	 
		case plugInModeHSLColor:		case plugInModeHSBColor:	case plugInModeIndexedColor:
		case plugInModeBitmap:			default:
			RR_ERRORSTRING("Unsupported image mode.  Plug-in cannot run. (ReadOriginalLine)");
		break;	
	}//switch (mode)
}//WriteResultLine


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
HANDLE hDllInstance;					//DLL instance handle

void main (void)						//need to have a main() even though we never use it
{	//unused
}//main

/*****************************************
*										 *
*	Code executed on DLL LoadLibrary.    *
*	Simply records the instance handle   *
*									     *
*****************************************/
extern BOOL WINAPI DllMain(HINSTANCE hModule, DWORD dwFunction, LPVOID);
extern BOOL WINAPI __stdcall DllMain(HINSTANCE hModule, DWORD dwFunction, LPVOID lpNot)
{
	hDllInstance = hModule;				// keep track of instance for this DLL
	lpNot = 0;							// suppresses a compiler warning
	switch (dwFunction)
	{
		case DLL_PROCESS_ATTACH:		// Message when DLL gets loaded
			hDllInstance = hModule;		// Record instance for later use
		break;
		case DLL_PROCESS_DETACH:		// Message when DLL is about to close
		break;							// Photoshop is quitting, presumably
		default:						// Does this ever happen?
		break;
	}
	return TRUE;
}//DllMain


/***********************************
*								   *
*	!!! PHOTOSHOP ENTRYPOINT !!!   *
*								   *
************************************/
void ENTRYPOINT (short PS_selector,
				 Edit_PSData *PS_Edit_Data,
				 long *PSdata,
				 short *PSresult)				
{	// This is the main plug-in entrypoint 
	OSErr	result = noErr;
	*PSdata = 0;
	switch (PS_selector)
	{	
		case filterSelectorAbout:					// Present the About... box to the user
			ErrorMessage(AboutBoxMessageCStr);		// string from PiPLData.h
		break;

		case filterSelectorParameters:				// Put up a dialog & fill in the parameters
		break;										// Don't allocate anything here - except the parameters handle

		case filterSelectorPrepare:					// Tell photoshop how much memory we need
			PS_Edit_Data->bufferSpace = 65536 * sizeof(RGBPixel) * 3;
			//PS_Edit_Data->maxSpace = 65536 * sizeof(RGBPixel) * 3;
		break;
		
		case filterSelectorStart:					// Begin the first pass on the image
			_RRStartupInitialization(PS_Edit_Data);	// This does all of the setup work 
													// dimensions, depth, bounding box, etc.
			result = MainUserEntry();				// CALL USER'S FUNCTION !!!
			_RRShutdownCleanup();					// we allocated
		break;
			
		case filterSelectorContinue:				// all subsequent processing passes (Photoshop 2.5 and before)
		break;										// shouldn't happen
		
		case filterSelectorFinish:					// Photoshop tells us to toss everything 
			_RRShutdownCleanup();					// we allocated, if we haven't already
		break;
			
		default:									// Some other funky selector we never heard of
			result = (OSErr)filterBadParameters;
		break;
	}
	*PSresult = result;								// propagate error message out
	
	//   Since we're using AdvanceState from within filterSelectorStart, all of filtering happens
	//   with just one call.  We don't have to do the multiple DoContinue from PS 2.5 and earlier.
}//PHOTOSHOP ENTRYPOINT

////////////////////////////////////////////////////////////////////////////////
//
//	Private function: _PhotoshopAdvanceState -- Photoshop's callback to read/write next bit of image
//
////////////////////////////////////////////////////////////////////////////////
OSErr	_PhotoshopAdvanceState()
{
	OSErr theerr = noErr;
	if ((gPrimaryImage.stuff->advanceState) == 0)
	{
		//note: This plug-in requires Photoshop 3.0.4 or later.
		return (OSErr)userCanceledErr;	//bad!
	}
	else
		theerr = (gPrimaryImage.stuff->advanceState)();
	return theerr;
}//_PhotoshopAdvanceState


////////////////////////////////////////////////////////////////////////////////
//
// Private function: _DoProgress -- Photoshop's callback for setting the progress bar
//
////////////////////////////////////////////////////////////////////////////////
void	_DoProgress(int32 value, int32 total)
{
	Edit_PSData *stuff = gPrimaryImage.stuff;
	if (!stuff) return;						  //don't have a stuff pointer	
	#define UpdateProgress2(done,total) ((*(stuff->progressProc)) (done, total))
		UpdateProgress2(value, total); 
	#undef UpdateProgress2
}//_DoProgress

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _CommandPeriod -- Photoshop's callback for polling the Cancel button or ESC 
//									   or CommandPeriod on the Mac
//
////////////////////////////////////////////////////////////////////////////////
Boolean	_CommandPeriod()
{
	Edit_PSData *stuff = gPrimaryImage.stuff;
	if (!stuff) return FALSE;				  //don't have a stuff pointer	
	#define TestAbort2() ((*stuff->abortProc) ())
	return TestAbort2();
	#undef TestAbort2
}//_CommandPeriod

////////////////////////////////////////////////////////////////////////////////
//
// TestAbortSetProgressBar	//0.0..1.0 range
//
////////////////////////////////////////////////////////////////////////////////
Boolean	TestAbortSetProgressBar(float percentage) //0..1
{
	float	position = (float)1000000000 * percentage;	//put progressbar position in range
	if (position < 0) position = 0;
	if (position > 1000000000)	position = 1000000000;
	//Note: Don't call this TOO often or you'll slow down processing
	_DoProgress((int32)position, (int32)1000000000);
	return _CommandPeriod();
}//DoProgressBarTestAbort


////////////////////////////////////////////////////////////////////////////////
//
// CreateTextFileToWrite
//
////////////////////////////////////////////////////////////////////////////////
ErrType	CreateTextFileToWrite(char *filename, File_ID *fileRef)
{
	ErrType	theerr = noErr;
	HANDLE	result;
	_FSID	*newFile = NULL;

	//create & open an empty file for writing
	result = CreateFile(filename, 	 					//address of file name
						GENERIC_READ | GENERIC_WRITE,	// access (read-write) mode 
						0,			 					// share mode 0 = none
	    				0,								// address of security descriptor 
						CREATE_ALWAYS,					// how to create 
						FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,	// file attributes 
						NULL 							// handle of file with attributes to copy  
					   );

	if (result == INVALID_HANDLE_VALUE)
	{
		switch (GetLastError())
		{
			case ERROR_ACCESS_DENIED:
				theerr = openErr;
			break;
		
			case EACCES://read only
				theerr = wrPermErr;
			break;
			case EINVAL://invalid parameters
				theerr = bdNamErr;
			break;
			case EMFILE://too many files
				theerr = tmfoErr;
			break;
			case ENOENT://file not found
				theerr = fnfErr;
			break;
			default:
				theerr = ioErr;
			break;
		}
	}
	if (theerr != noErr)
	{
		*fileRef = 0;	//no file
		RR_ERRORSTRING("Cannot open text file. (CreateTextFileToWrite)");
		return theerr;
	}

	newFile = CreateAPointer(1, sizeof(_FSID));
	newFile->fref = result;
	newFile->read = FALSE;		//write!!
	newFile->open = TRUE;
	newFile->nextFile = gOpenFileList;
	gOpenFileList = newFile;

	//return new write-only image to user
	*fileRef = (File_ID)((int32)newFile);	//hide structure from user
	return noErr;
}//CreateTextFileToWrite

////////////////////////////////////////////////////////////////////////////////
//
// OpenTextFileToRead
//
////////////////////////////////////////////////////////////////////////////////
ErrType	OpenTextFileToRead(char *filename, File_ID *fileRef)
{
	ErrType	theerr = noErr;
	HANDLE	result;
	_FSID	*newFile = NULL;

	result = CreateFile((LPCTSTR)filename, 	 	//address of file name
					    GENERIC_READ,			// access (read-only) mode 
					    0,			 			// share mode 0 = none
		    		    0,						// address of security descriptor 
					    OPEN_EXISTING,			// how to create 
					    FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,	// file attributes 
					    NULL 					// handle of file with attributes to copy  
					   );

	if (result == INVALID_HANDLE_VALUE)
	{
		switch (GetLastError())
		{
			case ERROR_ACCESS_DENIED:
				theerr = openErr;
			break;
		
			case EACCES://read only
				theerr = wrPermErr;
			break;
			case EINVAL://invalid parameters
				theerr = bdNamErr;
			break;
			case EMFILE://too many files
				theerr = tmfoErr;
			break;
			case ENOENT://file not found
				theerr = fnfErr;
			break;
			default:
				theerr = ioErr;
			break;
		}
	}
	if (theerr != noErr)
	{
		*fileRef = 0;	//no files here
		return theerr;
	}

	newFile = CreateAPointer(1, sizeof(_FSID));
	newFile->fref = result;
	newFile->read = TRUE;
	newFile->open = TRUE;
	newFile->nextFile = gOpenFileList;
	gOpenFileList = newFile;

	//return new read-only image to user
	*fileRef = (File_ID)((int32)newFile);	//hide structure from user
	return noErr;
}//OpenTextFileToRead

////////////////////////////////////////////////////////////////////////////////
//
// CloseTextFile
//
////////////////////////////////////////////////////////////////////////////////
void	CloseTextFile(File_ID fileRef)
{
	BOOL	success;
	_FSID	*currFile = (_FSID *)fileRef;
	if (currFile == NULL) return;	//??? bad file
	if (!(currFile->open))
	{
		RR_ERRORSTRING("Text file has already been closed. (CloseTextFile)");
		return;
	}
	success = CloseHandle(currFile->fref);
	//possible error is ENXIO == invalid handle
	if (!success)
	{
		RR_ERRORSTRING("Bad File_ID when closing text file. (CloseTextFile)");
		return;
	}
	currFile->open = FALSE;	//not going to delete the list, just mark off the entry
}//CloseTextFile

////////////////////////////////////////////////////////////////////////////////
// _CloseAllTextFiles
////////////////////////////////////////////////////////////////////////////////
void	_CloseAllTextFiles()
{
	_FSID *fileList = gOpenFileList;
	//Traverse linked list, close all files and dispose of pointers
	while (fileList != NULL)
	{
		_FSID *currFile = fileList;
		if (currFile->open)
			CloseTextFile((File_ID)((int32)currFile->fref));	//close each file
		fileList = currFile->nextFile;						//advance to next file
		DisposeAPointer(currFile);							//dispose memory for last file
	}
	gOpenFileList = NULL;
}//_CloseAllTextFiles

///////////////////////////////////////////////////////////////////
//	_ReadCH fetches a character from the input stream
///////////////////////////////////////////////////////////////////
ErrType	_ReadCH(HANDLE fref, char *inch, Boolean *eoln, Boolean *eof)
{
	DWORD	count = 1;
	ErrType	err = noErr;
	BOOL	success = ReadFile(fref,	// handle of file to read 
							   inch,	// address of buffer that receives data  
							   1,		// number of bytes to read 
							   &count,	// address of number of bytes actually read 
							   NULL 	// address of structure for data 
							   );	
	if (!success)
		err = ioErr;
	
	*eof = (Boolean)((err != noErr) || (count != 1));
	if (*eof)
	{
		err = noErr;		/* don't cause an error at eof */
		*inch = 26;			/* control-Z at eof */
	}
	*eoln = (*inch == 13);
	
	if (*eof)
		*eoln = TRUE;		/* end of file is also end of line */
	return(err);
}/*_ReadCH*/

///////////////////////////////////////////////////////////////////
//	_ReadCString knows about terminators, commas, tabs, CR, CRLF, etc.
///////////////////////////////////////////////////////////////////
ErrType	_ReadCString(HANDLE fref, char *astr, Boolean *eoln, Boolean *eof, long MaxChars)
{
	Boolean	hasTerm = FALSE;
	char	inch;
	Boolean	myeof, myeoln;
	char 	*bstr;
	short	nchars = 0;
	OSErr	err;
	Boolean	firstchar = TRUE;
	
	bstr = CreateAPointer(1, MaxChars);
	bstr[0] = 0;
	
	do
	{
		err = _ReadCH(fref, &inch, &myeoln, &myeof);
		if (err)
		{
			memcpy(astr, bstr, nchars + 1);
			DisposeAPointer(bstr);
			return(err);
		}
		if ((inch != 10) && (inch != 12) && (inch != 0))
		{
			if ((inch == 13) || (inch == 9) || myeoln || myeof || (inch == ',') || (inch == ' '))
				hasTerm = TRUE;
			else
			{
				// Got a character to add
				bstr[nchars] = inch;
				nchars++;
				bstr[nchars] = 0;
			}
			firstchar = FALSE; // don't eat spaces any longer
		}
		if (nchars > MaxChars - 2) // force a break if string gets too long
			hasTerm = TRUE;
		
	} while (!hasTerm && (err == noErr));
	
	// Just walked across a terminator, time to leave
	*eoln = myeoln;
	*eof = myeof;
	
	memcpy(astr, bstr, nchars + 1);
	DisposeAPointer(bstr);
	return(err);
} //_ReadCString


////////////////////////////////////////////////////////////////////////////////
//
// ReadANumber
//
////////////////////////////////////////////////////////////////////////////////
ErrType	ReadANumber(File_ID fileRef, float *value)
{
	Boolean	eof = FALSE, eoln = FALSE;
	const	int	MAX_CHARS = 256;
	char	*tempstring = NULL;
	ErrType	theerr = noErr;
	float	fetchedvalue = 0.0;
	_FSID	*currFile = (_FSID *)fileRef;
	if (currFile == NULL) return fnfErr;
	if (!(currFile->open)) return fnfErr;
	tempstring = CreateAPointer(1, MAX_CHARS);

	theerr = _ReadCString(currFile->fref, tempstring, &eoln, &eof, MAX_CHARS);
	//we'll ignore eoln.  EOF will precipitate an error unless there is data in the string

	if (strlen(tempstring) != 0)			//got something
	{
		theerr = noErr;						//override any error - we got some data
		fetchedvalue = (float)atof(tempstring);
	}
	else if ((eof) && (theerr == noErr))	//let user know we're at EOF 
		theerr = eofErr;
	if (tempstring)
		DisposeAPointer(tempstring);
	*value = fetchedvalue;					//pass value back out
	return theerr;
}//ReadANumber

////////////////////////////////////////////////////////////////////////////////
//
// WriteANumber
//		will always write a NewLine afterwards
////////////////////////////////////////////////////////////////////////////////
ErrType	WriteANumber(File_ID fileRef, float value)
{
	DWORD	count, bytesToWrite;
	BOOL	success;
	char	*tempstring = NULL;
	char	CRLF[3] = {13, 10, 0};
	const	int	MAX_CHARS = 256;
	ErrType	theerr = noErr;
	_FSID	*currFile = (_FSID *)fileRef;
	if (currFile == NULL) return fnfErr;
	if (!(currFile->open)) return fnfErr;

	tempstring = CreateAPointerClear(1, MAX_CHARS);
#ifdef _CRT_INSECURE_DEPRECATE				//if deprecate macro is defined, then should use _gcvt_s version
	(void)_gcvt_s(tempstring, MAX_CHARS, value, 5);
#else
	_gcvt(value, 5, tempstring);			//there is no _s (secure) version.  Use the original
#endif

	count = bytesToWrite = strlen(tempstring);

	success = WriteFile(currFile->fref,		// handle of file to write to 
					    (void *)tempstring,	// address of data to write to file 
					    bytesToWrite,		// number of bytes to write 
					    (LPDWORD)&count,	// address of number of bytes written 
					    NULL 				// addr. of structure needed for overlapped I/O  
					   );
	if (!success)
	{
		theerr = ioErr;			//default unless overridden
		switch (GetLastError())
		{
			case EBADF:
				theerr = ioErr;
			break;
			case ENOSPC:
				theerr = dskFulErr;
			break;
		}
	}
	if ((count == 0) && (theerr == noErr))	//failed silently
		theerr = ioErr;
	if (!theerr)	//append a CRLF
	{	
		bytesToWrite = strlen(CRLF);
		success = WriteFile(currFile->fref,	// handle of file to write to 
					    (void *)CRLF,		// address of data to write to file 
					    bytesToWrite,		// number of bytes to write 
					    (LPDWORD)&count,	// address of number of bytes written 
					    NULL 				// addr. of structure needed for overlapped I/O  
					   );

	}
	if (tempstring)
		DisposeAPointer(tempstring);
	return theerr;
}//WriteANumber


////////////////////////////////////////////////////////////////////////////////
//	REFERENCE IMAGE APIs
////////////////////////////////////////////////////////////////////////////////

#define MAX_FILENAME_PATH 1024
#define REFERENCE_IMAGE_NAME "Reference Image.xzy"

typedef HANDLE FILEptr;

typedef struct
{
	unsigned short xsize, ysize, depth;
} _diskImageData;

typedef struct _fileData					//private opaque data structure
{
	FILEptr	fref;							//file reference for file itself
	HANDLE	fileMapHandle;					//Handle for mapping
	_diskImageData	dimensions;				//height, width, depth
	Boolean	isStatic;						//meaning "save it on disk upon a close" or "we'll delete it"
	DWORD	imageSize;						//total number of bytes for image
	LPVOID	baseaddr;						//address of memory mapped file
	char	cStrName[MAX_FILENAME_PATH];	//name of file.  If no prefix, use temppath
	struct _fileData *next;					//doubly linked list of open files
	struct _fileData *prev;					//doubly linked list of open files
} _fileData, *_fcpptr;

_fileData *gOpenImageFileList = NULL;		//fileData for all open images
_fileData *gRefImage = NULL;				//fileData for reference image, if open

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _MakeStrictPathFileName
//
////////////////////////////////////////////////////////////////////////////////
void _MakeStrictPathFileName(char *cStringPath, long buffersize);
void _MakeStrictPathFileName(char *cStringPath, long buffersize)
{
	if (cStringPath == NULL)
		return;	//there are bigger problems with this name than the path

	if ((cStringPath[1] != ':') && (cStringPath[2] !='\\')) //no drive letter
	{
		char	temppath[MAX_FILENAME_PATH];
		GetTempPath(MAX_FILENAME_PATH, temppath);
		strcat(temppath, cStringPath);
		strcpy(cStringPath, temppath);
	}
}//_MakeStrictPathFileName

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _MakeNamedImage
//
////////////////////////////////////////////////////////////////////////////////
ErrType _MakeNamedImage(char *cStringName, Boolean isStatic, unsigned short x, unsigned short y, unsigned short depth, _fcpptr *fcp);
ErrType _MakeNamedImage(char *cStringName, Boolean isStatic, unsigned short x, unsigned short y, unsigned short depth, _fcpptr *fcp)
{
	OSErr	theerr = noErr;
	char	path[MAX_FILENAME_PATH];
	FILEptr fref = NULL;
	DWORD	bytewidth;
	DWORD	imagesize;
	_fcpptr	newfcp;
	HANDLE	fileMapHandle = NULL;
	LPVOID	baseaddr = NULL;
	DWORD	tcount;
	_diskImageData	sData;
	BOOL	success;

	//allocate a fcp record
	newfcp = (_fcpptr)CreateAPointerClear(sizeof(_fileData), 1);
	if (!newfcp)	return memFullErr;

	//populate record with useful data
	strcpy(path, cStringName);
	_MakeStrictPathFileName(path, MAX_FILENAME_PATH); //put it in the temp directory

	bytewidth = (unsigned long)x * (unsigned long)depth;
	imagesize = (unsigned long)y * bytewidth + sizeof(_diskImageData);	//include small header in file
	strcpy(newfcp->cStrName, path);
	newfcp->dimensions.xsize = x;
	newfcp->dimensions.ysize = y;
	newfcp->dimensions.depth = depth;
	newfcp->isStatic = isStatic;
	newfcp->imageSize = imagesize;
	
	//create the file either as a static file (won't delete) or temporary
	if (isStatic)	//static means it is going to be on disk...
		fref = CreateFile(path, 
						  GENERIC_WRITE | GENERIC_READ,	// access (read-write) mode 
						  0,			 	// share mode 0 = none
		    			  0,			// address of security descriptor 
						  CREATE_ALWAYS,	// how to create 
						  FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,	// file attributes 
						  NULL 			// handle of file with attributes to copy  
						 );
	else	//bias the performance for a temporary file, might not hit the disk
		fref = CreateFile(path, 
						  GENERIC_WRITE | GENERIC_READ,	// access (read-write) mode 
						  0,			 	// share mode 0 = none
		    			  0,			// address of security descriptor 
						  CREATE_ALWAYS,	// how to create 
						  FILE_ATTRIBUTE_TEMPORARY | FILE_FLAG_SEQUENTIAL_SCAN,	// file attributes 
						  NULL 			// handle of file with attributes to copy  
						 );

	if (fref == INVALID_HANDLE_VALUE)
	{
		switch (GetLastError())
		{
			case EACCES:												//read-only ?!
			case ERROR_ACCESS_DENIED:	theerr = wrPermErr;		break;	//access denied
			case EINVAL:				theerr = bdNamErr;		break;  //invalid parameters
			case EMFILE:				theerr = tmfoErr;		break;  //too many files
			case ENOENT:				theerr = fnfErr;		break;	//file not found
			default:					theerr = ioErr;			break;	//something else is wrong					
		}
		DisposeAPointer(newfcp);
		return theerr;
	}

	newfcp->fref = fref;

	//Write the header to the file
	sData.xsize = x;
	sData.ysize = y;
	sData.depth = depth;
	success = WriteFile(fref,	// handle of file to write 
					   &sData,		// address of buffer that holds data  
					   sizeof(_diskImageData),	// number of bytes to write 
					   &tcount,		// address of number of bytes actually written 
					   NULL 		// address of structure for data 
					   );	
	if (!success)
	{
		DisposeAPointer(newfcp);
		return ioErr;	//can't write to the file.
	}
	if (tcount != sizeof(_diskImageData))
	{
		DisposeAPointer(newfcp);
		return ioErr;
	}

	//try to memory map this file
	if (imagesize < 512 * 1024 * 1024 + 6) //512MB + header limit
	{
		fileMapHandle = CreateFileMapping(  fref,
											NULL,
											PAGE_READWRITE,
											0,
											imagesize,
											NULL);
		baseaddr = MapViewOfFile(	fileMapHandle,
									FILE_MAP_WRITE,
									0,	//start at 0
									0,
									0);	//whole file
		if (!baseaddr) // didn't map, oops.
		{
			CloseHandle(fileMapHandle);
			fileMapHandle = NULL;
		}
		theerr = noErr;
	}
	newfcp->baseaddr = baseaddr;
	newfcp->fileMapHandle = fileMapHandle;
	newfcp->next = newfcp->prev = NULL;

	//the code will use classic read/writes to access the data if the file mapping did not work
	//or if the image was simply too big.
	*fcp = newfcp;
	//add open file to linked list
	if (!gOpenImageFileList)
		gOpenImageFileList = newfcp;
	else
	{
		//append to front of gOpenImageFileList
		newfcp->next = gOpenImageFileList;
		gOpenImageFileList->prev = newfcp;
		gOpenImageFileList = newfcp;
	}

	return theerr;
}//_MakeNamedImage

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _OpenNamedImage
//
////////////////////////////////////////////////////////////////////////////////
ErrType _OpenNamedImage(char *cStringName, _fcpptr *fcp)	//always going to be the reference image
{
	OSErr	theerr = noErr;
	char	path[MAX_FILENAME_PATH];
	FILEptr fref = NULL;
	_diskImageData sData;
	DWORD	bytewidth;
	DWORD	imagesize;
	_fileData	*newfcp;
	HANDLE	fileMapHandle = NULL;
	LPVOID	baseaddr = NULL;
	BOOL	success;
	DWORD	tcount;
	unsigned short x, y, depth;

	newfcp = (_fileData *)CreateAPointerClear(sizeof(_fileData), 1);
	if (!newfcp)	return memFullErr;

	strcpy(path, cStringName);
	_MakeStrictPathFileName(path, MAX_FILENAME_PATH); //put it in the temp directory

	strcpy(newfcp->cStrName, path);
	newfcp->isStatic = TRUE;
	
	fref = CreateFile(path, 
					  GENERIC_WRITE | GENERIC_READ,	// access (read-write) mode 
					  0,			 	// share mode 0 = none
		    		  0,			// address of security descriptor 
					  OPEN_EXISTING,	// how to create 
					  FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,	// file attributes 
					  NULL 			// handle of file with attributes to copy  
					 );
	if (fref == INVALID_HANDLE_VALUE)
	{
		switch (GetLastError())
		{
			case EACCES:												//read-only ?!
			case ERROR_ACCESS_DENIED:	theerr = wrPermErr;		break;	//access denied
			case EINVAL:				theerr = bdNamErr;		break;  //invalid parameters
			case EMFILE:				theerr = tmfoErr;		break;  //too many files
			case ENOENT:				theerr = fnfErr;		break;	//file not found
			default:					theerr = ioErr;			break;	//something else is wrong					
		}
		{
			DisposeAPointer(newfcp);
			return theerr;
		}
	}

	newfcp->fref = fref;

	//read the file header to determine size
	success = ReadFile(fref,	// handle of file to read 
					   &sData,		// address of buffer that receives data  
					   sizeof(_diskImageData),	// number of bytes to read 
					   &tcount,		// address of number of bytes actually read 
					   NULL 		// address of structure for data 
					   );	
	if (!success)
	{
		DisposeAPointer(newfcp);
		return ioErr;	//can't read the file.
	}

	x = sData.xsize;
	y = sData.ysize;
	depth = sData.depth;

	//calculate file dimensions
	bytewidth = (unsigned long)x * (unsigned long)depth;
	imagesize = (unsigned long)y * bytewidth + sizeof(_diskImageData);	//include small header in file
	newfcp->dimensions.xsize = x;
	newfcp->dimensions.ysize = y;
	newfcp->dimensions.depth = depth;
	newfcp->imageSize = imagesize;

	//try to memory map this file
	if (imagesize < 512 * 1024 * 1024 + 6) //512MB + header limit
	{
		fileMapHandle = CreateFileMapping(  fref,
											NULL,
											PAGE_READWRITE,
											0,
											imagesize,
											NULL);
		baseaddr = MapViewOfFile(	fileMapHandle,
									FILE_MAP_WRITE,
									0,	//start at 0
									0,
									0);	//whole file
		if (!baseaddr) // didn't map, oops.
		{
			CloseHandle(fileMapHandle);
			fileMapHandle = NULL;
		}
		theerr = noErr;
	}
	newfcp->baseaddr = baseaddr;
	newfcp->fileMapHandle = fileMapHandle;
	newfcp->next = newfcp->prev = NULL;

	//the code will use classic read/writes to access the data if the file mapping did not work
	//or if the image was simply too big.

	*fcp = newfcp;
	//add open file to linked list
	if (!gOpenImageFileList)
		gOpenImageFileList = newfcp;
	else
	{
		//append to front of gOpenImageFileList
		newfcp->next = gOpenImageFileList;
		gOpenImageFileList->prev = newfcp;
		gOpenImageFileList = newfcp;
	}
	return theerr;
}//_OpenNamedImage

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _CloseNamedImage
//
////////////////////////////////////////////////////////////////////////////////
void	_CloseNamedImage(_fcpptr fcp)
{
	OSErr	theerr = noErr;
	BOOL 		success;
	int	result;

	if (!fcp)
	{
		RR_ERRORSTRING("Try to close a NULL Image_ID (_CloseNamedImage)");
		return;
	}
	if (fcp->baseaddr)
	{	//end memory mapping
		(void)FlushViewOfFile(fcp->baseaddr, fcp->imageSize);
		(void)UnmapViewOfFile(fcp->baseaddr);
		CloseHandle(fcp->fileMapHandle);
	}
	success = CloseHandle(fcp->fref);
	if (!success)
	{
		long	dummy = GetLastError();
		theerr = rfNumErr;
	}
	if (!fcp->isStatic)
	{
		result = remove(fcp->cStrName);
		if (result == -1)
			switch (GetLastError())
			{
				case EACCES:	theerr = fLckdErr;	break;
				case ENOENT:	theerr = fnfErr;	break;
				default:		theerr = ioErr;		break;
			}
		//actually, we don't care about an error here.
	}
	///remove open file from list
	if (!gOpenImageFileList)
	{
		RR_ERRORSTRING("Try to close an image when there are none (_CloseNamedImage)");
	}
	else
	{
		if (fcp->prev == NULL)	//must be at head of the list
		{
			gOpenImageFileList = fcp->next;
			if (fcp->next)
				gOpenImageFileList->prev = NULL;
		}
		else if (fcp->next == NULL)//delete at end of list
		{
			if (fcp->prev)
				fcp->prev->next = NULL;
		}
		else //delete in middle, standard doubly linked-list delete
		{
			if (fcp->prev)
				fcp->prev->next = fcp->next;
			if (fcp->next)
				fcp->next->prev = fcp->prev;
		}
	}
	DisposeAPointer(fcp);

	//return(theerr);
}//_CloseNamedImage

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _TossNamedImage
//
////////////////////////////////////////////////////////////////////////////////
void	_TossNamedImage(_fcpptr fcp)
{
	if (!fcp)
		return;
	fcp->isStatic = FALSE;	//make it so that close will purge it!
	_CloseNamedImage(fcp);
}//_TossNamedImage

////////////////////////////////////////////////////////////////////////////////
//
//	InitializeRefImage 
//		literally, create a new reference image with these dimensions x RGBPixel
////////////////////////////////////////////////////////////////////////////////
ErrType	InitializeRefImage(long width, long height)					//initializes new Reference image
{
	_fcpptr	newfcp;
	OSErr	theerr = noErr;
	if (gRefImage)	_TossNamedImage(gRefImage);
	gRefImage = NULL;
	theerr = _MakeNamedImage(REFERENCE_IMAGE_NAME, TRUE, (unsigned short)width, (unsigned short)height, sizeof(RGBPixel), &newfcp);
	if (theerr == noErr)
		gRefImage = newfcp;
	return theerr;
}//InitializeRefImage

////////////////////////////////////////////////////////////////////////////////
//
//	GetRefDimensions
//
////////////////////////////////////////////////////////////////////////////////
ErrType	GetRefDimensions(long *width, long *height)						//get Reference image dimensions
{
	_fcpptr	newfcp;
	OSErr	theerr = noErr;
	if (!gRefImage)
	{
		theerr = _OpenNamedImage(REFERENCE_IMAGE_NAME, &newfcp);
		if (theerr == noErr)
			gRefImage = newfcp;
	}
	if (theerr == noErr)
	{
		*width = gRefImage->dimensions.xsize;
		*height = gRefImage->dimensions.ysize;
	}
	else
	{
		*width = 0;
		*height = 0;
	}
	return theerr;
}//GetRefDimensions

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _GetColumn
//
////////////////////////////////////////////////////////////////////////////////
void	_GetColumn(_fcpptr fcp, long x, void *line);
void	_GetColumn(_fcpptr fcp, long x, void *line)
{
	OSErr	theerr = noErr;
	unsigned long	height, width, depth, bytewidth;
	unsigned long	srcaddr;
	long	result;
	BOOL	success;
	long	tcount;
	unsigned char *lineptr = line;
	long	y;

	width = fcp->dimensions.xsize;
	if ((x < 0) || (x >= (long)width))
	{
		RR_ERRORSTRING("X column out of range for this image (_GetColumn)");
		return;
	}

	height = fcp->dimensions.ysize;
	depth = fcp->dimensions.depth;
	bytewidth = depth * width;

	if (fcp->baseaddr) //memcpy version
		for (y = 0; y < (long)height; y++)
		{
			srcaddr = (unsigned long)(fcp->baseaddr) + sizeof(_diskImageData) + 
						x * depth + bytewidth * (unsigned long)y;
			memcpy(lineptr, (LPVOID)srcaddr, depth);
			lineptr = &lineptr[depth];	//advance by depth bytes
		}//for y
	else //file read version
		for (y = 0; y < (long)height; y++)
		{
			//set file position
			srcaddr = sizeof(_diskImageData) + x * depth + bytewidth * (unsigned long)y;
			result = SetFilePointer(fcp->fref,	// handle of file 
									srcaddr,	// number of bytes to move file pointer 
									NULL,		// address of high-order word of distance to move  
									FILE_BEGIN 	// how to move 
								   );
			if (result == -1)
			{
				switch(GetLastError())
				{
					case EBADF:		theerr = rfNumErr;	break; //bad refnum
					case EINVAL:	theerr = posErr;	break; //bad file position
					default:		theerr = ioErr;		break; //some other problem
				}
				return; //not actually passing the error out
			}
			//read depth bytes
			tcount = depth;
			success = ReadFile(fcp->fref,	// handle of file to read 
							   lineptr,	// address of buffer that receives data  
							   depth,	// number of bytes to read 
							   &tcount,	// address of number of bytes actually read 
							   NULL 	// address of structure for data 
							   );	
			if (tcount != depth)
				theerr = ioErr;
			if (!success)
				theerr = ioErr;
			if (theerr != noErr)	
				return;
			//advance lineptr pointer
			lineptr = &lineptr[depth];	//advance by depth bytes
		}//for y
}//_GetColumn

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _PutColumn
//
////////////////////////////////////////////////////////////////////////////////
void	_PutColumn(_fcpptr fcp, long x, void *line);
void	_PutColumn(_fcpptr fcp, long x, void *line)
{
	OSErr	theerr = noErr;
	unsigned long	height, width, depth, bytewidth;
	unsigned long	srcaddr;
	long	result;
	BOOL	success;
	long	tcount;
	unsigned char *lineptr = line;
	long	y;

	width = fcp->dimensions.xsize;
	if ((x < 0) || (x >= (long)width))
	{
		RR_ERRORSTRING("X column out of range for this image (_PutColumn)");
		return;
	}

	height = fcp->dimensions.ysize;
	depth = fcp->dimensions.depth;
	bytewidth = depth * width;

	if (fcp->baseaddr) //memcpy version
		for (y = 0; y < (long)height; y++)
		{
			srcaddr = (unsigned long)(fcp->baseaddr) + sizeof(_diskImageData) + 
						x * depth + bytewidth * (unsigned long)y;
			memcpy((LPVOID)srcaddr, lineptr, depth);
			lineptr = &lineptr[depth];	//advance by depth bytes
		}//for y
	else //file read version
		for (y = 0; y < (long)height; y++)
		{
			//set file position
			srcaddr = sizeof(_diskImageData) + x * depth + bytewidth * (unsigned long)y;
			result = SetFilePointer(fcp->fref,	// handle of file 
									srcaddr,	// number of bytes to move file pointer 
									NULL,		// address of high-order word of distance to move  
									FILE_BEGIN 	// how to move 
								   );
			if (result == -1)
			{
				switch(GetLastError())
				{
					case EBADF:		theerr = rfNumErr;	break; //bad refnum
					case EINVAL:	theerr = posErr;	break; //bad file position
					default:		theerr = ioErr;		break; //some other problem
				}
				return; //not actually passing the error out
			}
			//write depth bytes
			tcount = depth;
			success = WriteFile(fcp->fref,	// handle of file to write 
							   lineptr,	// address of buffer that holds data  
							   depth,	// number of bytes to write 
							   &tcount,	// address of number of bytes actually written 
							   NULL 	// address of structure for data 
							   );	
			if (tcount != depth)
				theerr = ioErr;
			if (!success)
				switch(GetLastError())
				{
					case EBADF:		theerr = ioErr;		break;
					case ENOSPC:	theerr = dskFulErr; break;
					default:		theerr = ioErr;		break;
				}
			if (theerr != noErr)	
				return;
			//advance lineptr pointer
			lineptr = &lineptr[depth];	//advance by depth bytes
		}//for y
}//_PutColumn

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _GetLine
//
////////////////////////////////////////////////////////////////////////////////
void	_GetLine(_fcpptr fcp, long y, void *line);
void	_GetLine(_fcpptr fcp, long y, void *line)
{
	OSErr	theerr = noErr;
	unsigned long	height, bytewidth;
	unsigned long	srcaddr;
	long	result;
	BOOL	success;
	long	tcount;

	height = fcp->dimensions.ysize;
	if ((y < 0) || (y >= (long)height))
	{
		RR_ERRORSTRING("Y line out of range for this image (_GetLine)");
		return;
	}
	bytewidth = (unsigned long)fcp->dimensions.xsize * (unsigned long)fcp->dimensions.depth; 
		
	if (fcp->baseaddr)
	{
		srcaddr = (unsigned long)(fcp->baseaddr) + sizeof(_diskImageData) + bytewidth * (unsigned long)y;
		memcpy(line, (LPVOID)srcaddr, bytewidth);
	}
	else
	{
		srcaddr = sizeof(_diskImageData) + bytewidth * (unsigned long)y;
		//Set File Position for this line
		result = SetFilePointer(fcp->fref,	// handle of file 
								srcaddr,	// number of bytes to move file pointer 
								NULL,	// address of high-order word of distance to move  
								FILE_BEGIN 	// how to move 
							   );
		if (result == -1)
		{
			switch(GetLastError())
			{
				case EBADF:		theerr = rfNumErr;	break; //bad refnum
				case EINVAL:	theerr = posErr;	break; //bad file position
				default:		theerr = ioErr;		break; //some other problem
			}
			return; //not actually passing the error out
		}

		//read data
		tcount = bytewidth;
		success = ReadFile(fcp->fref,	// handle of file to read 
						   line,	// address of buffer that receives data  
						   bytewidth,	// number of bytes to read 
						   &tcount,	// address of number of bytes actually read 
						   NULL 	// address of structure for data 
						   );	
		if (tcount != bytewidth)
			theerr = ioErr;
		if (!success)
			theerr = ioErr;
		//no way to pass error messages out
	}
}//_GetLine

////////////////////////////////////////////////////////////////////////////////
//
// Private function: _PutLine
//
////////////////////////////////////////////////////////////////////////////////
void	_PutLine(_fcpptr fcp, long y, void *line);
void	_PutLine(_fcpptr fcp, long y, void *line)
{
	OSErr	theerr = noErr;
	unsigned long	height, bytewidth;
	unsigned long	srcaddr;
	long	result;
	BOOL	success;
	long	tcount;

	height = fcp->dimensions.ysize;
	if ((y < 0) || (y >= (long)height))
	{
		RR_ERRORSTRING("Y line out of range for this image (_PutLine)");
		return;
	}
	bytewidth = (unsigned long)fcp->dimensions.xsize * (unsigned long)fcp->dimensions.depth; 
		
	if (fcp->baseaddr)
	{
		srcaddr = (unsigned long)(fcp->baseaddr) + sizeof(_diskImageData) + bytewidth * (unsigned long)y;
		memcpy((LPVOID)srcaddr, line, bytewidth);
	}
	else
	{
		srcaddr = sizeof(_diskImageData) + bytewidth * (unsigned long)y;
		//Set File Position for this line
		result = SetFilePointer(fcp->fref,	// handle of file 
								srcaddr,	// number of bytes to move file pointer 
								NULL,	// address of high-order word of distance to move  
								FILE_BEGIN 	// how to move 
							   );
		if (result == -1)
		{
			switch(GetLastError())
			{
				case EBADF:		theerr = rfNumErr;	break; //bad refnum
				case EINVAL:	theerr = posErr;	break; //bad file position
				default:		theerr = ioErr;		break; //some other problem
			}
			return; //not actually passing the error out
		}

		//write data
		tcount = bytewidth;
		success = WriteFile(fcp->fref,	// handle of file to write 
						   line,	// address of buffer that provides data  
						   bytewidth,	// number of bytes to write 
						   &tcount,	// address of number of bytes actually written 
						   NULL 	// address of structure for data 
						   );	
		if (tcount != bytewidth)
			theerr = ioErr;
		if (!success)
			switch(GetLastError())
			{
				case EBADF:		theerr = ioErr;		break;
				case ENOSPC:	theerr = dskFulErr; break;
				default:		theerr = ioErr;		break;
			}
		//no way to pass error messages out
	}
}//_PutLine

////////////////////////////////////////////////////////////////////////////////
//
//	ReadRefImageLine
//
////////////////////////////////////////////////////////////////////////////////
void	ReadRefImageLine(long y, RGBPixel *line)	//read and write work with replacement
{
	_fcpptr	newfcp;
	OSErr	theerr = noErr;

	if (!gRefImage)
	{
		theerr = _OpenNamedImage(REFERENCE_IMAGE_NAME, &newfcp);
		if (theerr == noErr)
			gRefImage = newfcp;
	}
	if (theerr != noErr)
	{
		RR_ERRORSTRING("There is no reference image (ReadRefImageLine)");
		return;
	}
	_GetLine(gRefImage, y, line);
}//ReadRefImageLine

////////////////////////////////////////////////////////////////////////////////
//
//	WriteRefImageLine
//
////////////////////////////////////////////////////////////////////////////////
void	WriteRefImageLine(long y, RGBPixel *line)
{
	_fcpptr	newfcp;
	OSErr	theerr = noErr;

	if (!gRefImage)
	{
		theerr = _OpenNamedImage(REFERENCE_IMAGE_NAME, &newfcp);
		if (theerr == noErr)
			gRefImage = newfcp;
	}
	if (theerr != noErr)
	{
		RR_ERRORSTRING("There is no reference image (WriteRefImageLine)");
		return;
	}
	_PutLine(gRefImage, y, line);
}//WriteRefImageLine

////////////////////////////////////////////////////////////////////////////////
//
//	MakeTempoararyImage
//
////////////////////////////////////////////////////////////////////////////////
ErrType	MakeTemporaryImage(long width, long height, long bytesperpix,		//Allocate a temporary image workspace
							char *uniqueName, Image_ID *imageID)			// in whatever bytedepth you like
{
	OSErr	theerr = noErr;
	_fcpptr	newfcp = NULL;
	*imageID = 0;
	theerr = _MakeNamedImage(uniqueName, FALSE, (unsigned short)width, (unsigned short)height, (unsigned short)bytesperpix, &newfcp);
	if (theerr == noErr)
		*imageID = (int32)newfcp;
	return theerr;
}//MakeTemporaryImage

////////////////////////////////////////////////////////////////////////////////
//
//	ReadTempImageLine
//
////////////////////////////////////////////////////////////////////////////////
void	ReadTempImageLine(Image_ID imageID, long y, void *line)			//read and write work with replacement
{
	if (imageID)
		_GetLine((_fcpptr)imageID, y, line);
	else
	{
		RR_ERRORSTRING("imageID is 0.  (ReadTempImageLine)");
		return;
	}
}//ReadTempImageLine

////////////////////////////////////////////////////////////////////////////////
//
//	WriteTempImageLine
//
////////////////////////////////////////////////////////////////////////////////
void	WriteTempImageLine(Image_ID imageID, long y, void *line)
{
	if (imageID)
		_PutLine((_fcpptr)imageID, y, line);
	else
	{
		RR_ERRORSTRING("imageID is 0.  (WriteTempImageLine)");
		return;
	}
}//WriteTempImageLine

////////////////////////////////////////////////////////////////////////////////
//
//	ReadTempImageColumn
//
////////////////////////////////////////////////////////////////////////////////
void	ReadTempImageColumn(Image_ID imageID, long x, void *line)		//!!must allocate line to be large enough
{
	if (imageID)
		_GetColumn((_fcpptr)imageID, x, line);
	else
	{
		RR_ERRORSTRING("imageID is 0.  (ReadTempImageColumn)");
		return;
	}
}//ReadTempImageColumn

////////////////////////////////////////////////////////////////////////////////
//
//	WriteTempImageColumn
//
////////////////////////////////////////////////////////////////////////////////
void	WriteTempImageColumn(Image_ID imageID, long x, void *line)			//!!must allocate line to be large enough
{
	if (imageID)
		_PutColumn((_fcpptr)imageID, x, line);
	else
	{
		RR_ERRORSTRING("imageID is 0.  (WriteTempImageColumn)");
		return;
	}
}//WriteTempImageColumn

////////////////////////////////////////////////////////////////////////////////
//
//	Private Function: _RRStartupInitialization
//
////////////////////////////////////////////////////////////////////////////////
void	_RRStartupInitialization(Edit_PSData *PSStuff)
{
	long	pc = 0;
	long	bpc = 0;
	//initialize memory system
	gPointerBlockListHead = NULL;
	gCurrentBlock = NULL;
	gBlockPointerCount = 0;
	//initialize original image global data
	gPrimaryImage.mode = PSStuff->imageMode; // 0 = Photoshop 3.0, 1..4 = Photoshop 2.5
	gPrimaryImage.result = noErr;				//for error propagation
	gPrimaryImage.stuff = PSStuff;	
	gPrimaryImage.width = PSStuff->imageSize.h;
	gPrimaryImage.height = PSStuff->imageSize.v;
	gPrimaryImage.width = PSStuff->imageSize.h;
	gPrimaryImage.height = PSStuff->imageSize.v;
	gPrimaryImage.bitdepth = PSStuff->depth;			//either 1, 8, 16, or 32 bits per channel
	switch (gPrimaryImage.mode)
	{
		case plugInModeBitmap:
			pc = 1; bpc = 1;
		break;
		case plugInModeGrayScale: 		//supported
			pc = 1; bpc = 8;
			if (PSStuff->depth == 32)	//FIXES A BUG IN PHOTOSHOP
			{
				gPrimaryImage.mode = plugInModeGray32;
				bpc = 32;
			}
		break;
		case plugInModeIndexedColor:
			pc = 1; bpc = 8;
		break;
		case plugInModeLabColor:		//supported via transform
		case plugInModeHSLColor:
		case plugInModeHSBColor:
		case plugInModeRGBColor:		//supported
			pc = 3; bpc = 8;
		break;
		case plugInModeCMYKColor: 
			pc = 4; bpc = 8;
		case plugInModeMultichannel: 
		case plugInModeDuotone:	 
			pc = gPrimaryImage.stuff->planes;	//# channels in image
			bpc = 8;
		break;
		case plugInModeGray16:			//supported
			pc = 1; bpc = 16;
		break;
		case plugInModeLab48:			//supported via transform
		case plugInModeRGB48:			//supported
			pc = 3; bpc = 16;
		break;
		case plugInModeCMYK64:
			pc = 4; bpc = 16;
		break;
		case plugInModeDuotone16: 
		case plugInModeDeepMultichannel:
			pc = gPrimaryImage.stuff->planes;	//# channels in image
			bpc = 16;
		break;
		case plugInModeRGB96:			//supported
			pc = 3; bpc = 32;
		break;
		case plugInModeGray32:      	//supported
			pc = 1; bpc = 32;
		break;

		default:
			pc = 0;
			bpc = 0;
		return;
	}
	gPrimaryImage.planeCount = pc;				//# channels in image
	//Initialize rect / AdvanceState parameters
	gPrimaryImage.stuff->inRect.top =		gPrimaryImage.stuff->outRect.top =		0;
	gPrimaryImage.stuff->inRect.bottom =	gPrimaryImage.stuff->outRect.bottom =	1;	//depending on line#0 for first read
	gPrimaryImage.stuff->inRect.left =		gPrimaryImage.stuff->outRect.left =		0;
	gPrimaryImage.stuff->inRect.right =		gPrimaryImage.stuff->outRect.right =	(short)gPrimaryImage.width;
	if (gPrimaryImage.stuff->haveMask)
	{
		gPrimaryImage.stuff->maskRect.top =		0;
		gPrimaryImage.stuff->maskRect.bottom =	1;
		gPrimaryImage.stuff->maskRect.left =	0;
		gPrimaryImage.stuff->maskRect.right = (short)gPrimaryImage.width;
	}
	else
	{
		gPrimaryImage.stuff->maskRect.top =		0;
		gPrimaryImage.stuff->maskRect.bottom =	0;
		gPrimaryImage.stuff->maskRect.left =	0;
		gPrimaryImage.stuff->maskRect.right =	0;
	}
	//Plane (channel) selection
	gPrimaryImage.stuff->outLoPlane = gPrimaryImage.stuff->inLoPlane = 0;
	gPrimaryImage.stuff->outHiPlane = gPrimaryImage.stuff->inHiPlane = (short)gPrimaryImage.planeCount - 1;
	gPrimaryImage.stuff->inputRate = gPrimaryImage.stuff->maskRate = 0x00010000; //1.0 FIXED POINT
	gPrimaryImage.stuff->maskPadding = -1;										 //plugInWantsEdgeReplication

	//Cached Writes -- to work with AdvanceState and decouple reads from writes
	gPrimaryImage.cachedWritePtr = CreateAPointer(gPrimaryImage.width, sizeof(RGBPixel));
	gPrimaryImage.cachedWriteLineNum = 0;			//line#0 is queued up to be read and then cached
	gPrimaryImage.cachedLineLength = 0;
	gPrimaryImage.firstTimeAdvanceState = TRUE;
	gPrimaryImage.hasCachedData = FALSE;

	//initialize text file list
	gOpenFileList = NULL;
	//initialize access to Reference image
	gRefImage = NULL;
	//initialize temporary image file list
	gOpenImageFileList = NULL;
}//_RRStartupInitialization

////////////////////////////////////////////////////////////////////////////////
//
//	Private function: _RRShutdownCleanup()
//
////////////////////////////////////////////////////////////////////////////////
void	_RRShutdownCleanup()
{
	//PREVENT PROBLEMS WITH MULTIPLE CALLS
	if (gPrimaryImage.cachedWritePtr == NULL)
		return;
	//finish up any remaining writes
	if (gPrimaryImage.hasCachedData)
	{
		//request for outData buffer for the remaining cached line
		gPrimaryImage.stuff->outRect.top = (short)gPrimaryImage.cachedWriteLineNum;			 //set line number for cached data
		gPrimaryImage.stuff->outRect.bottom = (short)gPrimaryImage.cachedWriteLineNum + 1; 
		(void)_PhotoshopAdvanceState();														 //push an advancestate to clear out what was already there
		memcpy(gPrimaryImage.stuff->outData, gPrimaryImage.cachedWritePtr, 
											 gPrimaryImage.stuff->outRowBytes);				 //copy data into buffer for final write
	}//flush cached data
	DisposeAPointer(gPrimaryImage.cachedWritePtr);
	gPrimaryImage.cachedWritePtr = NULL;
	//auto-close Reference image

	if (gRefImage)
		_CloseNamedImage(gRefImage);
	//auto-dispose of temporary image files
	while (gOpenImageFileList)
		_TossNamedImage(gOpenImageFileList);
	
	//auto-close opened text files **warning on open text files
	_CloseAllTextFiles();
	//dispose of original image global data
	//tell Photoshop that we're really done, not requesting ANY data and to write out whatever is left in outData
	gPrimaryImage.stuff->inRect.top =		gPrimaryImage.stuff->outRect.top =		gPrimaryImage.stuff->maskRect.top =		0;
	gPrimaryImage.stuff->inRect.bottom =	gPrimaryImage.stuff->outRect.bottom =	gPrimaryImage.stuff->maskRect.bottom =	0;
	gPrimaryImage.stuff->inRect.left =		gPrimaryImage.stuff->outRect.left =		gPrimaryImage.stuff->maskRect.left =	0;
	gPrimaryImage.stuff->inRect.right =		gPrimaryImage.stuff->outRect.right =	gPrimaryImage.stuff->maskRect.right =	0;
}//__RRShutdownCleanup()