/*
	File: PhotoshopShell.h

	portions Copyright (c) 1990-1, Thomas Knoll.
	portions Copyright (c) 1992-6, Adobe Systems Incorporated.
	portions Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved

	This file incorporates some portions of versions 3.0.4, 4.0, 
	a few lines of 5.0, and the plug-in modes of 9.0 from Photoshop's Filter 
	module interface. It also incorporates the API described in Russ&Russ, 2007

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or
	distributed for a fee.
*/
#define _CRT_SECURE_NO_DEPRECATE							//don't want silly messages about strcat & strcpy
#include <string.h>											//so this overrides deprecate messages for userCode.c only

#include <windows.h>


#ifndef __PSSHELL__
#define __PSSHELL__

#define RRDEBUG 1

#ifdef RRDEBUG
	#define RR_ERRORSTRING(a) ErrorMessage(a)	//show error message
#else
	#define RR_ERRORSTRING(a) {}				//don't show error message
#endif


/*********************************************************************************/
/* Operation selectors, calling states for Photoshop */ 
#define filterSelectorAbout 	 0
#define filterSelectorParameters 1
#define filterSelectorPrepare	 2
#define filterSelectorStart 	 3
#define filterSelectorContinue	 4
#define filterSelectorFinish	 5

/*********************************************************************************/
/* IMAGE Modes for Photoshop */
#define plugInModeBitmap	   0
#define plugInModeGrayScale    1		//supported
#define plugInModeIndexedColor 2
#define plugInModeRGBColor	   3		//supported
#define plugInModeCMYKColor    4
#define plugInModeHSLColor	   5
#define plugInModeHSBColor	   6
#define plugInModeMultichannel 7
#define plugInModeDuotone	   8
#define plugInModeLabColor	   9		//supported via transform
#define plugInModeGray16	  10		//supported
#define plugInModeRGB48		  11		//supported
#define plugInModeLab48		  12		//supported via transform
#define plugInModeCMYK64	  13
#define plugInModeDeepMultichannel 14
#define plugInModeDuotone16   15
#define plugInModeRGB96		  16		//supported
#define plugInModeGray32      17		//supported

/*********************************************************************************/
/* TYPES used throughout */
typedef signed long  int32;
typedef signed short int16;
typedef signed char  int8;
typedef unsigned long  unsigned32;
typedef unsigned short unsigned16;
typedef unsigned char  unsigned8;
typedef unsigned8	uint8;
typedef unsigned16	uint16;
typedef unsigned32	uint32;
typedef unsigned char BYTE;
typedef LPSTR	Ptr;
typedef LPSTR	*Handle;
typedef BYTE Boolean;
typedef long Fixed;	//16.16 fixed point
typedef DWORD	OSType;
typedef unsigned char Str255[256]; //first byte length of string - Pascal string
/*********************************************************************************/
/* PHOTOSHOP ERROR CODES - Mac codes used throughout Mac and PC versions of Photoshop */
typedef WORD	OSErr;
#define noErr 0
#define userCanceledErr		(-128)
#define coercedParamErr		2
#define readErr				(-19)
#define writErr				(-20)
#define openErr				(-23)
#define	dirFulErr			(-33)		// directory full
#define dskFulErr			(-34)
#define	nsvErr				(-35)		// no such volume
#define ioErr				(-36)
#define	bdNamErr			(-37)		// bad name
#define	fnOpnErr			(-38)		// file not open
#define eofErr				(-39)
#define	posErr				(-40)		// cannot position before start
#define	mFulErr				(-41)		// memory full (open)
#define	tmfoErr				(-42)		// too many files open
#define fnfErr				(-43)
#define	wPrErr				(-44)		// disc is write-protected
#define	fLckdErr			(-45)		// file is locked
#define vLckdErr			(-46)
#define	fBsyErr				(-47)		// file is busy
#define	opWrErr				(-49)		// file already open for write
#define paramErr			(-50)
#define	rfNumErr			(-51)		// bad fRefNum
#define wrPermErr			(-61)		// write permissions doesn't allow writing
#define memFullErr			(-108)
#define nilHandleErr		(-109)
#define memWZErr			(-111)
#define filterBadParameters -30100		// "a problem with the filter module interface"
#define filterBadMode		-30101		// "the filter module does not support <mode> images"

/*********************************************************************************/
/* BOOLEAN values and other useful constants */
#define nil		   NULL
#ifndef TRUE
	#define TRUE 1
#endif
#ifndef FALSE
	#define FALSE 0
#endif
#ifndef true
	#define true       TRUE
#endif
#ifndef false
	#define false      FALSE
#endif

/*********************************************************************************/
/* PHOTOSHOP STRUCTURES */
typedef struct Point
{
  short v;
  short h;
} Point;

typedef struct Rect
{
  short top;
  short left;
  short bottom;
  short right;
} Rect;

typedef struct
{
	WORD red;
	WORD green;
	WORD blue;
} RGBColor;
/*********************************************************************************/
/* CALLBACKS for Photoshop v. 3.0.4 and later */
#ifdef __cplusplus
extern "C"
{
#endif
	typedef  Boolean (*TestAbortProc) (void);
	typedef  void (*ProgressProc) (int32 done, int32 total);
	typedef  OSErr (*AdvanceStateProc) (void);
#ifdef __cplusplus
}
#endif

#if WIN32
#pragma pack(push,4)
#endif

/********************************************************************************/
/* ABBREVIATED Photoshop FilterRecord structure */
typedef struct 
{
	int32			reserved0;
	TestAbortProc	abortProc;		/* The plug-in module may call this no-argument
									   BOOLEAN function (using Pascal calling
									   conventions) several times a second during long
									   operations to allow the user to abort the operation.
									   If it returns TRUE, the operation should be aborted
									   (and userCanceledErr returned). */
	ProgressProc	progressProc;	/* The plug-in module may call this two-argument
									   procedure periodically to update a progress
									   indicator.  The first parameter is the number
									   of operations completed; the second is the total
									   number of operations. */
	int32			reserved1;
	Point			imageSize;		/* Size of image, up to 30,000 x 30,000 */
	int16			planes; 		/* Samples per pixel -- can include opacity and 
									   selection planes.  Also called "channels" */
	Rect			filterRect; 	/* Rectangle to filter */
	RGBColor		background; 	/* Current background color */
	RGBColor		foreground; 	/* Current foreground color */
	int32			maxSpace;		/* Maximum possible total of data and buffer space */
	int32			bufferSpace;	/* If the plug-in filter needs to allocate
									   large internal buffers, the filterSelectorPrepare
									   routine should set this field to the number
									   of bytes the filterSelectorStart routine is
									   planning to allocate.  Relocatable blocks should
									   be used if possible. */
	Rect			inRect; 		/* Requested input rectangle. Must be a subset of
									   the image's bounding rectangle. */
	int16			inLoPlane;		/* First requested input plane (channel) */
	int16			inHiPlane;		/* Last requested input plane (channel) */
	Rect			outRect;		/* Requested output rectangle. Must be a subset of
									   filterRect. */
	int16			outLoPlane; 	/* First requested output plane (channel) */
	int16			outHiPlane; 	/* Last requested output plane (channel) */
	void		   *inData; 		/* Pointer to input rectangle. If more than one
									   plane (channel) was requested, the data is interleaved. */
	int32			inRowBytes; 	/* Offset between input rows */
	void		   *outData;		/* Pointer to output rectangle. If more than one
									   plane (channel) was requested, the data is interleaved. */
	int32			outRowBytes;	/* Offset between output rows */
	Boolean			isFloating; 	/* Set to true if the selection is floating, 
									   early version of Photoshop only, unused with layers */
	Boolean			haveMask;		/* Set to true if there is a selection mask */
	Boolean			autoMask;		/* If there is a mask, and the selection is not
									   floating, the plug-in can change this field to
									   false to turn off auto-masking. */
	Rect			maskRect;		/* Requested mask rectangle.  Must be a subset of
									   filterRect. Should only be used if haveMask is
									   true. */
	void		   *maskData;		/* Pointer to (read only) mask data. */
	int32			maskRowBytes;	/* Offset between mask rows */
	int32			reserved2[4];
	int16			imageMode;		/* Image mode */
	Fixed			imageHRes;		/* Pixels per inch */
	Fixed			imageVRes;		/* Pixels per inch */
	Point			floatCoord; 	/* Top left coordinate of selection */
	Point			wholeSize;		/* Size of image selection is floating over */
	int32			reserved3[10];
	void 		   *platformData;	/* Platform specific information. */
	int32			reserved4[8];
	AdvanceStateProc advanceState;	/* Advance from start to continue or
									   continue to continue. */
	int16			reserved5[6];	
	int16			maskPadding;	/* Padding instructions for the mask. */
	int16			reserved6; 
	Fixed			inputRate;		/* Input sample rate. */
	Fixed			maskRate;		/* Mask sample rate. */
	int32			reserved7[21];
	int32			depth;			/* bit depth per channel (1,8,16, 32) */
	char			reserved8[66]; 	/* Set to zero */
} Edit_PSData, *Edit_PSDataPtr;

#if WIN32
#pragma pack(pop)
#endif

#if PRAGMA_STRUCT_ALIGN
#pragma options align=reset
#endif

/********************************************************************************/
//////////////////////////////////////////////////////////////////////////////////
//	//	//	//	//	//	//	//	//	R&R API	//	//	//	//	//	//	//	//	//	//
//////////////////////////////////////////////////////////////////////////////////
/********************************************************************************/
//R&R TYPES

typedef int32 Image_ID;	//opaque structure for temp image data
typedef int32 File_ID;	//opaque structure for file data
typedef int16 ErrType;	//See list of Photoshop errors above

//R&R STRUCTURES
typedef struct
{
	float red;
	float green;
	float blue;		//standard format for all images.  value is USUALLY 0..255
} RGBPixel;

typedef struct
{
	float real;
	float imag;			//standard format for FFT function.
} FFTelem;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//R&R FUNCTION PROTOTYPES
	//Original image (RGBPixel) - Photoshop's current layer or selection.  											
	//Note: you CANNOT read what you write in the primary image
void	GetOriginalDimensions(long *width, long *height);					//Get dimensions of this image
void	ReadOriginalLine(long y, RGBPixel *line);							//read from image
void	WriteResultLine(long y, RGBPixel *line);							//write to image

	//Reference image (RGBPixel)- our archived static image for two image operations												
	//Note: you CAN read what you write in the reference image
ErrType	InitializeRefImage(long width, long height);						//initializes new Reference image
ErrType	GetRefDimensions(long *width, long *height);						//get Reference image dimensions
void	ReadRefImageLine(long y, RGBPixel *line);							//read and write work with replacement
void	WriteRefImageLine(long y, RGBPixel *line);

	//Temporary image(s) (any format) - temporary block of memory, but accessed with Read/Write functions
	//Note: you CAN read what you write in any temporary image
ErrType	MakeTemporaryImage(long width, long height, long bytesperpix,		//Allocate a temporary image workspace
							char *uniquenName, Image_ID *imageID);			// in whatever bytedepth you like
void	ReadTempImageLine(Image_ID imageID, long y, void *line);			//read and write work with replacement
void	WriteTempImageLine(Image_ID imageID, long y, void *line);
void	ReadTempImageColumn(Image_ID imageID, long x, void *line);			//!!must allocate line to be large enough
void	WriteTempImageColumn(Image_ID imageID, long x, void *line);			//!!must allocate line to be large enough

	//Utility functions, memory functions are protected!  Don't use malloc!
void	ErrorMessage(char *message);										//put a c-string up in a dialog
void   *CreateAPointer(long length, long size);								//allocates (length*size). NULL if it fails
void   *CreateAPointerClear(long length, long size);						//allocates and fills with zeros.  NULL if it fails
void	DisposeAPointer(void *pointer);										//deallocates a pointer, protected against NULL

	//Text file access
ErrType	CreateTextFileToWrite(char *filename, File_ID *fileRef);
ErrType	OpenTextFileToRead(char *filename, File_ID *fileRef);
void	CloseTextFile(File_ID fileRef);
ErrType	ReadANumber(File_ID fileRef, float *value);
ErrType	WriteANumber(File_ID fileRef, float value);

	//progress bar / command period test
//Push Photoshop's progress bar and test to see if user cancels (ESC/CommandPeriod)
Boolean	TestAbortSetProgressBar(float percentage);			//goes from 0.0->1.0, set the bar anyplace you want
															//and, see if the user wants to cancel
//USER ENTRY POINT
ErrType	MainUserEntry();	//you get called.  return one of the errors at the end.   hopefully noErr.

#endif //__PSSHELL__
