
/* Photoshop plug-in example */
/*
	Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved.

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
	distributed for a fee.
*/
/* Problem 2.1.10.11 Sharpening Filter */

#include "PhotoshopShell.h"

typedef struct	// holds spherical LAB values for each pixel
{	float L, A, B;
}LABPixel;

typedef	RGBPixel *RGBLinePtr; // pointer for RGB pixel values
typedef LABPixel *LABLinePtr; // and for LAB values


void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline);
void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline)
{
	long	x; // pixel address across each line
	for (x = 0; x < width; x++)
	{	// convert RGB to spherical LAB values
		LABline[x].L = (RGBline[x].red + RGBline[x].green + RGBline[x].blue)/3;
		LABline[x].A = (RGBline[x].red - RGBline[x].green);
		LABline[x].B = (RGBline[x].red + RGBline[x].green)/2 - RGBline[x].blue;
	}
}

void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline);
void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline)
{
	long	x;
	for (x = 0; x < width; x++)
	{	// convert from spherical LAB values to RGB
		RGBline[x].red =   LABline[x].L + LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].green = LABline[x].L - LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].blue =  LABline[x].L - 2 * LABline[x].B/3;
	}
}

#define RAD 4 // neighborhood size = 9x9 (central pixel plus 4 in all directions)
#define size 2*RAD+1 // kernel size

///////////////////////////////////
//
//	Main entry for user code
//
///////////////////////////////////
ErrType	MainUserEntry()	
{
	ErrType result = noErr;						//accumulate any errors that happened
	ErrType theerr = noErr;
	long	height, width;						//pixel dimensions of image
	long	x, y;								//loop counters to traverse image
	long	i, j, h, k;							//loop counters for neighborhood
	float	kernel[size][size];					//3x3 weights
	float	sum, diff, final, divisor = 0;		//used to compute kernel product
	float	pixel[size][size];					//pixel values
	RGBLinePtr RGBline[size];					//memory buffer - multiple lines of RGB
	LABLinePtr LABline[size];					//and LAB pixel values
	RGBPixel *ResultRGB;						//result line to write back
	LABPixel *ResultLAB;						//intermediate result in LAB
	long	fref;								//file ref number for filter values
	float	sigma = (float)2.1;					//std dev. of Gaussian blur (read from file)
	float	delta = (float)3.5;					//threshold for applying difference
	float	factor = (float)1.5;				//amplify usm difference

	GetOriginalDimensions(&width, &height);		//find out how big the image is

	for (i = 0; i < size; i++)					//initialize pointers to null
	{	
		RGBline[i] = NULL;
		LABline[i] = NULL;
	}
	ResultRGB = NULL;
	ResultLAB = NULL;
	
	for (i = 0; i < size; i++)					//allocate memory for multiple lines of pixels
	{	
		RGBline[i] = CreateAPointer(width, sizeof(RGBPixel));
		LABline[i] = CreateAPointer(width, sizeof(LABPixel));	
	}//Allocate memory for 2*RAD+1 rows
	ResultRGB = CreateAPointer(width, sizeof(RGBPixel));//result values in RGB
	ResultLAB = CreateAPointer(width, sizeof(LABPixel));//and in LAB

	if (OpenTextFileToRead("C:\\GBlur.txt", &fref) != noErr)//look for the input file
	{	
		result = fnfErr;
		goto mainexit;
	}
	else
	{	// read sigma for blur from file
		theerr = ReadANumber(fref, &sigma);
		CloseTextFile(fref);
		if (theerr != noErr)
			{	
				result = theerr;
				goto mainexit;
			}	
	}
	
	//calculate the Gaussian kernel values and their sum
	for (i = 0; i < size; i++)
		for (j = 0; j < size; j++)
		{	
			float	r, denom;
			denom = 2 * sigma * sigma;
			h = i - RAD;
			k = j - RAD;
			r = (h*h+k*k) / denom;
			kernel[i][j]=(float)exp(-r)/(denom*(float)3.1415926535897);
			divisor += kernel[i][j];
		}

	for (y = 0; y < height; y++)// traverse the image line by line
	{	
		if (TestAbortSetProgressBar((float)y/(float)height))
		{	//See if user canceled and update progress bar
			result = userCanceledErr;
			goto mainexit;
		}
		for (i = 0; i < size; i++) //read in multiple lines of original values
		{
			h = y + i - RAD;//vertical offset
			if (h<0) ReadOriginalLine(0, RGBline[i]); // boundary protection at top
			else if (h>=height) ReadOriginalLine(height-1, RGBline[i]);// and bottom
			else ReadOriginalLine(h, RGBline[i]);
		}

		for (i = 0; i < size; i ++)
			ConvertLineRGBtoLAB(width, RGBline[i], LABline[i]);// convert RGB->LAB
			
		for (x = 0; x < width; x++)// traverse each line of the image
		{
			for (i = 0; i < size; i++)// build array of neighborhood pixels
				for (j = 0; j < size; j++)
				{
					k = x + j - RAD;//horizontal offset
					if (k<0) pixel[i][j] = LABline[i][0].L;//boundary protection at left
					else if (k>=width) pixel[i][j] = LABline[i][width-1].L;//and right
					else pixel[i][j] = LABline[i][k].L;
				}
			// apply the convolution kernel to the pixels
			sum = 0;
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++)
					sum += kernel[i][j] * pixel[i][j];
			sum /= divisor;
			diff = factor * (pixel[RAD][RAD] - sum);// amplify the result
			if (fabs(diff) > delta) // test absolute magnitude of difference
			{	
				final = pixel[RAD][RAD] + diff; // and add to original
				if (final < 0) final = 0;
				if (final > 255) final = 255;
			}
			else final = pixel[RAD][RAD];// otherwise leave it alone
			ResultLAB[x].L = final;
			ResultLAB[x].A = LABline[1][x].A;//keep original color values
			ResultLAB[x].B = LABline[1][x].B;
		}//for x
		ConvertLineLABtoRGB(width, ResultLAB, ResultRGB);//convert back to RGB
		WriteResultLine(y, ResultRGB);// and write to final image
	}//for y
	
mainexit:	//clean up everything we allocated
	for (i = 0; i < size; i++)
	{	if (RGBline[i]) DisposeAPointer(RGBline[i]);
		if (LABline[i]) DisposeAPointer(LABline[i]);
	}
	if (ResultRGB) DisposeAPointer(ResultRGB);
	if (ResultLAB) DisposeAPointer(ResultLAB);

	return result;											//pass any errors back
}//MainUserEntry

