
/* Photoshop plug-in example */
/*
	Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved.

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
	distributed for a fee.
*/
/* Problem 2.1.10.15 Conditional Processing*/

#include "PhotoshopShell.h"

typedef struct	// holds spherical LAB values for each pixel
{	float L, A, B;
}LABPixel;

typedef	RGBPixel *RGBLinePtr; // pointer for RGB pixel values
typedef LABPixel *LABLinePtr; // and for LAB values


void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline);
void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline)
{	
	long	x; // pixel addresses across each line
	for (x = 0; x < width; x++)
	{	// convert RGB to spherical LAB values
		LABline[x].L = (RGBline[x].red + RGBline[x].green + RGBline[x].blue)/3;
		LABline[x].A = (RGBline[x].red - RGBline[x].green);
		LABline[x].B = (RGBline[x].red + RGBline[x].green)/2 - RGBline[x].blue;
	}
}

void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline);
void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline)
{
	long	x;
	for (x = 0; x < width; x++)
	{	// convert from spherical LAB values to RGB
		RGBline[x].red =   LABline[x].L + LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].green = LABline[x].L - LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].blue =  LABline[x].L - 2 * LABline[x].B/3;
	}
}

#define RAD 4 // neighborhood size = 9x9 (central pixel plus 4 in all directions)
#define size 2*RAD+1 // kernel size

///////////////////////////////////
//
//	Main entry for user code
//
///////////////////////////////////
ErrType	MainUserEntry()	
{	ErrType result = noErr;						//accumulate any errors that happened
	ErrType theerr = noErr;
	long	height, width;						//pixel dimensions of image
	long	x, y;								//loop counters to traverse image
	long	i, j, h, k;							//counters to traverse neighborhood
	float	kernel[size][size];					//3x3 weights
	float	gauss, divisor = 0;					//calculations for kernel
	float	vert, horz, sobel, laplac, fraction, final;// summations for filter application
	float	pixel[size][size];					//pixel values
	RGBLinePtr RGBline[size];					//memory buffer - multiple lines of original pixels
	LABLinePtr LABline[size];					//and converted LAB values
	RGBPixel *ResultRGB;						//final result 
	LABPixel *ResultLAB;						//intermediate result in LAB
	long	fref;								//file ref number for filter values
	float	sigma = (float)2.1;					//std dev. of Gaussian blur (read from file)

	GetOriginalDimensions(&width, &height);		//find out how big the image is

	for (i = 0; i < size; i++)					//initialize pointers to NULL
	{											//(simplifies cleanup below)
		RGBline[i] = NULL;
		LABline[i] = NULL;
	}
	ResultRGB = NULL;
	ResultLAB = NULL;
	
	for (i = 0; i < size; i++)					//allocate memory for 2*RAD+1 rows
	{	RGBline[i] = CreateAPointer(width, sizeof(RGBPixel));
		LABline[i] = CreateAPointer(width, sizeof(LABPixel));	
	}
	ResultRGB = CreateAPointer(width, sizeof(RGBPixel));//and for results
	ResultLAB = CreateAPointer(width, sizeof(LABPixel));

	if (OpenTextFileToRead("C:\\GBlur.txt", &fref) != noErr)//find the input file
	{
		result = fnfErr;
		goto mainexit;
	}
	else
	{	// read sigma for blur from file
		theerr = ReadANumber(fref, &sigma);
		CloseTextFile(fref);
		if (theerr != noErr)
			{	
				result = theerr;
				goto mainexit;
			}	
	}
	
	//calculate the Gaussian kernel and its sum
	for (i = 0; i < size; i++)
		for (j = 0; j < size; j++)
		{
			float	r, denom;
			denom = 2 * sigma * sigma;
			h = i - RAD;
			k = j - RAD;
			r = (h*h+k*k) / denom;
			kernel[i][j]=(float)exp(-r)/(denom*(float)3.14159265359897);
			divisor += kernel[i][j];
		}
		
	for (y = 0; y < height; y++)//traverse the image line by line
	{
		if (TestAbortSetProgressBar((float)y/(float)height))
		{	//See if user canceled and update progress bar
			result = userCanceledErr;
			goto mainexit;
		}
		for (i = 0; i < size; i++)// read multiple lines of original pixels
		{
			h = y + i - RAD;//vertical offset
			if (h<0) ReadOriginalLine(0, RGBline[i]); // boundary protection at top
			else if (h>=height) ReadOriginalLine(height-1, RGBline[i]);// and bottom
			else ReadOriginalLine(h, RGBline[i]);
		}

		for (i = 0; i < size; i ++)
			ConvertLineRGBtoLAB(width, RGBline[i], LABline[i]);//convert RGB->spherical LAB
		
		for (x = 0; x < width; x++)// traverse each line pixel by pixel
		{
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++) // read neighborhood pixels
				{
					k = x + j - RAD;// horizontal offset
					if (k<0) pixel[i][j] = LABline[i][0].L;// boundary protection at left
					else if (k>=width) pixel[i][j] = LABline[i][width-1].L;// and right
					else pixel[i][j] = LABline[i][k].L;
				}
				
			// apply the Gaussian convolution kernel to the pixels
			gauss = 0;
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++)
					gauss += kernel[i][j] * pixel[i][j];
			gauss /= divisor;
			// apply the Laplacian filter to the pixels
			laplac = 9 * pixel[RAD][RAD];
			for (i = -1; i <= +1; i++)
				for (j = -1; j <= +1; j++)
				{	
					if ((i != 0) || (j != 0)) // not for the center
						laplac -= pixel[RAD+i][RAD+j];
				}
				
			// apply the Sobel filter to the pixels
			vert = 0; horz = 0;
			for (i = -1; i <= +1; i++)
			{	
				vert += pixel[RAD-1][RAD+i] - pixel[RAD+1][RAD+i];
				horz += pixel[RAD+i][RAD-1] - pixel[RAD+i][RAD+1];
			}
			sobel = (float)sqrt(vert*vert+horz*horz);
			fraction = sobel / 255; // arbitrary scaling of the Sobel result
			
			//combine the multiple filter results
			final = fraction * laplac + (1 - fraction) * gauss;
			
			//protect against over- or underflow
			if (final < 0) final = 0;
			if (final > 255) final = 255;
			
			ResultLAB[x].L = final;
			ResultLAB[x].A = LABline[1][x].A;// keep the original color values
			ResultLAB[x].B = LABline[1][x].B;
		}//for x
		ConvertLineLABtoRGB(width, ResultLAB, ResultRGB);// convert back to RGB
		WriteResultLine(y, ResultRGB);// and rewrite the results
	}//for y
	
mainexit:	//clean up everything we allocated
	for (i = 0; i < size; i++)
	{	if (RGBline[i]) DisposeAPointer(RGBline[i]);
		if (LABline[i]) DisposeAPointer(LABline[i]);
	}
	if (ResultRGB) DisposeAPointer(ResultRGB);
	if (ResultLAB) DisposeAPointer(ResultLAB);

	return result;											//pass any errors back
}//MainUserEntry

