
/* Photoshop plug-in example */
/*
	Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved.

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
	distributed for a fee.
*/
/* Problem 2.1.10.6 Gaussian Filter */

#include "PhotoshopShell.h"

typedef struct	// define a struct to hold pixel LAB values 
{	float L, A, B;
}LABPixel;

typedef	RGBPixel *RGBLinePtr;	//pointers to lines of pixels with RGB values
typedef LABPixel *LABLinePtr;	//same for LAB values


void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline);
void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline)
{	long	x;
	for (x = 0; x < width; x++)
	{	// convert RGB to spherical LAB values
		LABline[x].L = (RGBline[x].red + RGBline[x].green + RGBline[x].blue)/3;
		LABline[x].A = (RGBline[x].red - RGBline[x].green);
		LABline[x].B = (RGBline[x].red + RGBline[x].green)/2 - RGBline[x].blue;
	}
}

void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline);
void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline)
{	long	x;
	for (x = 0; x < width; x++)
	{	// convert from spherical LAB values to RGB
		RGBline[x].red =   LABline[x].L + LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].green = LABline[x].L - LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].blue =  LABline[x].L - 2 * LABline[x].B/3;
	}
}

#define RAD 3 // neighborhood size = 7x7 (central pixel plus radius of 3 in all directions)
#define size 2*RAD+1

///////////////////////////////////
//
//	Main entry for user code
//
///////////////////////////////////
ErrType	MainUserEntry()	
{	ErrType result = noErr;						//accumulate any errors that happened
	ErrType theerr = noErr;
	long	height, width;						//pixel dimensions of image
	long	x, y;								//loop counters to traverse image
	long	i, j, h, k;							//loop counters for kernel addressing
	float	kernel[size][size];					//array of kernel weights
	float	sum, divisor, offset = 0;			//calculation of sum for the neighborhood
	float	pixel[size][size];					//pixel values
	RGBLinePtr RGBline[size];					//memory buffer - multiple lines of RGB pixels
	LABLinePtr LABline[size];					//and the same for spherical LAB values
	RGBPixel *ResultRGB;						//final result to write back
	LABPixel *ResultLAB;						//result in LAB values
	long	fref;								//file ref number for filter values
	float	sigma = 1.5;						// std dev. of Gaussian blur (read from file)

	GetOriginalDimensions(&width, &height);		//find out how big the image is

	for (i = 0; i < size; i++)					//initialize pointers to NULL
	{	
		RGBline[i] = NULL;						//(not essential but simpliflies cleanup below)
		LABline[i] = NULL;
	}
	ResultRGB = NULL;
	ResultLAB = NULL;
	
	for (i = 0; i < size; i++)
	{	
		RGBline[i] = CreateAPointer(width, sizeof(RGBPixel));	//allocate memory for lines of pixels
		LABline[i] = CreateAPointer(width, sizeof(LABPixel));	
	}//Allocate memory for 2*RAD+1 rows
	ResultRGB = CreateAPointer(width, sizeof(RGBPixel));
	ResultLAB = CreateAPointer(width, sizeof(LABPixel));

	if (OpenTextFileToRead("C:\\GBlur.txt", &fref) != noErr) //look for the text file with gauss std. dev.
	{
		result = fnfErr;
		goto mainexit;
	}
	else
	{	// read sigma for blur from file
		theerr = ReadANumber(fref, &sigma);
		CloseTextFile(fref);
		if (theerr != noErr)
			{	
				result = theerr;
				goto mainexit;
			}	
	}
	
	// create the Gaussian kernel and sum the values to normalize the result
	divisor = 0; 
	for (i = 0; i < size; i++)
		for (j = 0; j < size; j++)
		{	
			float	r, denom;
			denom = 2 * sigma * sigma;
			h = i - RAD;
			k = j - RAD;
			r = (h*h+k*k) / denom;
			kernel[i][j]=(float)exp(-r)/(denom*(float)3.1415926535897);
			divisor += kernel[i][j];
		}
		
	/*	//diagnostic routine to write the kernel out to a text file for verification
	{	// write the filter to a text file
		long fref = 0; // file ID number
		if (CreateTextFileToWrite("C:\\gauss.txt", &fref) == 0);//location to write file
		{	
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++)
					WriteANumber(fref, (float)kernel[i][j]/divisor);
			CloseTextFile(fref);
		}// if file was successfully opened
	}
	*/
	
	for (y = 0; y < height; y++)	// for every line of pixels
	{	
		if (TestAbortSetProgressBar((float)y/(float)height))
		{	//See if user canceled and update progress bar
			result = userCanceledErr;
			goto mainexit;
		}
		for (i = 0; i < size; i++) // read multiple lines
		{	h = y + i - RAD; //offset to lines above and below
			if (h<0) ReadOriginalLine(0, RGBline[i]); // boundary protection above
			else if (h>=height) ReadOriginalLine(height-1, RGBline[i]);// and below
			else ReadOriginalLine(h, RGBline[i]);
		}

		for (i = 0; i < size; i ++)
			ConvertLineRGBtoLAB(width, RGBline[i], LABline[i]); // convert RGB->spherical LAB
		
		for (x = 0; x < width; x++)	// for every pixel in the row
		{	
			for (i = 0; i < size; i++) // traverse the neighborhood
				for (j = 0; j < size; j++)
				{	
					k = x + j - RAD; // offset to pixels left and right
					if (k<0) pixel[i][j] = LABline[i][0].L; // boundary protection at left
					else if (k>=width) pixel[i][j] = LABline[i][width-1].L;// and right
					else pixel[i][j] = LABline[i][k].L;
				}
			// apply the convolution kernel to the pixels
			sum = 0;
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++)
					sum += kernel[i][j] * pixel[i][j];
			sum = (sum / divisor) + offset; // note - offset is zero in this case
			if (sum < 0) sum = 0; // over/underflow can't happen in this case
			if (sum > 255) sum = 255;//  but good practice to protect against it
			ResultLAB[x].L = sum;
			ResultLAB[x].A = LABline[RAD][x].A;// keep the color values from the original
			ResultLAB[x].B = LABline[RAD][x].B;
		}//for x
		
		ConvertLineLABtoRGB(width, ResultLAB, ResultRGB);// convert back to RGB
		WriteResultLine(y, ResultRGB);// and write results back to image
	}//for y
	
mainexit:	//clean up everything we allocated
	for (i = 0; i < size; i++)
	{	if (RGBline[i]) DisposeAPointer(RGBline[i]);
		if (LABline[i]) DisposeAPointer(LABline[i]);
	}
	if (ResultRGB) DisposeAPointer(ResultRGB);
	if (ResultLAB) DisposeAPointer(ResultLAB);

	return result;	//pass any errors back
}//MainUserEntry

