
/* Photoshop plug-in example */
/*
	Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved.

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
	distributed for a fee.
*/
/* Problem 2.3.4.1 Variance */

#include "PhotoshopShell.h"

typedef struct // to hold spherical LAB values for a pixel
{	float L, A, B;
}LABPixel;

typedef	RGBPixel *RGBLinePtr;// pointer to a row of RGB pixel values
typedef LABPixel *LABLinePtr;// pointer to a row of LAB pixel values

void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline);
void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline)
{
	long	x;
	for (x = 0; x < width; x++)
	{	// convert RGB to spherical LAB values
		LABline[x].L = (RGBline[x].red + RGBline[x].green + RGBline[x].blue)/3;
		LABline[x].A = (RGBline[x].red - RGBline[x].green);
		LABline[x].B = (RGBline[x].red + RGBline[x].green)/2 - RGBline[x].blue;
	}
}

void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline);
void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline)
{
	long	x;
	for (x = 0; x < width; x++)
	{	// convert from spherical LAB values to RGB
		RGBline[x].red =   LABline[x].L + LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].green = LABline[x].L - LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].blue =  LABline[x].L - 2 * LABline[x].B/3;
	}
}
#define RAD 2 // neighborhood size = 5x5 (central pixel plus radius of 2 in all directions)
#define size 2*RAD+1

///////////////////////////////////
//
//	Main entry for user code
//
///////////////////////////////////
ErrType	MainUserEntry()	
{
	ErrType result = noErr;									//accumulate any errors that happened
	long	height, width;									//pixel dimensions of image
	long	x, y;											//loop counters to traverse image
	long	i, j, h, k;										//loop counters for neighborhood
	float	min, max;										//for autoscaling
	float	sum1, sum2, mean, variance, stddev, count, temp;//to calculate variance
	float	pixel[size][size];								//pixel values in neighborhood
	RGBLinePtr RGBline[size];								//memory buffer - 5 lines
	LABLinePtr LABline[size];								//lines after conversion to LAB
	RGBPixel *ResultRGB;									//final result
	LABPixel *ResultLAB;									//intermediate result

	GetOriginalDimensions(&width, &height);					//find out how big the image is

	for (i = 0; i < size; i++)								//initialize pointers to null (easier cleanup)
	{
		RGBline[i] = NULL;
		LABline[i] = NULL;
	}
	ResultRGB = NULL;
	ResultLAB = NULL;
	
	for (i = 0; i < size; i++)
	{
		RGBline[i] = CreateAPointer(width, sizeof(RGBPixel));
		LABline[i] = CreateAPointer(width, sizeof(LABPixel));	//Allocate memory for multiple rows
	}
	ResultRGB = CreateAPointer(width, sizeof(RGBPixel));	//and for results
	ResultLAB = CreateAPointer(width, sizeof(LABPixel));

	min = (float)1e20; max = 0; // initialize limits for autoscaling
	for (y = 0; y < height; y++) // first pass - scan line by line to get min & max
	{
		if (TestAbortSetProgressBar((float)y/(float)height))
		{	//See if user canceled and update progress bar
			result = userCanceledErr;
			goto mainexit;
		}
		for (i = 0; i < size; i++) // read multiple lines
		{
			h = y + i - RAD;//vertical offset
			if (h<0) ReadOriginalLine(0, RGBline[i]); // boundary protection at top
			else if (h>=height) ReadOriginalLine(height-1, RGBline[i]);//and bottom
			else ReadOriginalLine(h, RGBline[i]);
		}
		for (i = 0; i < size; i ++)
			ConvertLineRGBtoLAB(width, RGBline[i], LABline[i]);//work with LAB values
		
		for (x = 0; x < width; x++)//for every pixel in the row
		{
			for (i = 0; i < size; i++)// get the neighbors
				for (j = 0; j < size; j++)
				{
					k = x + j - RAD;// horizontal offset
					if (k<0) pixel[i][j] = LABline[i][0].L;//boundary protection at left
					else if (k>=width) pixel[i][j] = LABline[i][width-1].L;//and right
					else pixel[i][j] = LABline[i][k].L;
				}
			// calculate the variance 
			sum1 = 0; sum2 = 0; count = 0;
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++)
				{	// eliminate corners for circular nborhood
					if (((i > 0) && (i < size-1)) || ((j > 0) && (j < size-1)))
					{
						sum1 += pixel[i][j];
						sum2 +=	pixel[i][j] * pixel[i][j];
						count += 1.0;
					}
				}
			mean = sum1/count;
			variance = (sum2/count) - (mean * mean);
			stddev = (float)sqrt(variance);// this is called a variance filter but it's really the standard deviation
			if (stddev < min) min = stddev;
			if (stddev > max) max = stddev;//now have limits for autoscaling...
		}//for x
	}//for y

	for (y = 0; y < height; y++) // do it all again but apply autoscaling and save results
	{
		if (TestAbortSetProgressBar((float)y/(float)height))
		{	
			result = userCanceledErr;
			goto mainexit;
		}
		for (i = 0; i < size; i++)
		{	h = y + i - RAD;
			if (h<0) ReadOriginalLine(0, RGBline[i]);
			else if (h>=height) ReadOriginalLine(height-1, RGBline[i]);
			else ReadOriginalLine(h, RGBline[i]);
		}
		for (i = 0; i < size; i ++)
			ConvertLineRGBtoLAB(width, RGBline[i], LABline[i]);
		
		for (x = 0; x < width; x++)
		{
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++)
				{
					k = x + j - RAD;
					if (k<0) pixel[i][j] = LABline[i][0].L;
					else if (k>=width) pixel[i][j] = LABline[i][width-1].L;
					else pixel[i][j] = LABline[i][k].L;
				}

			sum1 = 0; sum2 = 0; count = 0;
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++)
				{	if (((i > 0) && (i < size-1)) || ((j > 0) && (j < size-1)))
					{	sum1 += pixel[i][j];
						sum2 +=	pixel[i][j] * pixel[i][j];
						count += 1.0;
					}
				}
			mean = sum1/count;
			variance = (sum2/count) - (mean * mean);
			stddev = (float)sqrt(variance);
			temp = 255*(stddev-min)/(max-min);// autoscaling step
			ResultLAB[x].L = temp;//combine resulting brightness value
			ResultLAB[x].A = LABline[1][x].A;//with original pixel color values
			ResultLAB[x].B = LABline[1][x].B;
		}//for x
		ConvertLineLABtoRGB(width, ResultLAB, ResultRGB);//convert back to RGB
		WriteResultLine(y, ResultRGB);//and save
	}//for y
	
mainexit:	//clean up everything we allocated
	for (i = 0; i < size; i++)
	{	if (RGBline[i]) DisposeAPointer(RGBline[i]);
		if (LABline[i]) DisposeAPointer(LABline[i]);
	}
	if (ResultRGB) DisposeAPointer(ResultRGB);
	if (ResultLAB) DisposeAPointer(ResultLAB);

	return result;											//pass any errors back
}//MainUserEntry

