
/* Photoshop plug-in example */
/*
	Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved.

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
	distributed for a fee.
*/
/* Problem 2.3.4.8 Kuwahara (Max Likelihood) */

#include "PhotoshopShell.h"

typedef struct // holds spherical LAB pixel values
{	float L, A, B;
}LABPixel;

typedef	RGBPixel *RGBLinePtr; // line pointers for pixels
typedef LABPixel *LABLinePtr;

void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline);
void ConvertLineRGBtoLAB(long width, RGBPixel *RGBline, LABPixel *LABline)
{	long	x;
	for (x = 0; x < width; x++)
	{	// convert RGB to spherical LAB values
		LABline[x].L = (RGBline[x].red + RGBline[x].green + RGBline[x].blue)/3;
		LABline[x].A = (RGBline[x].red - RGBline[x].green);
		LABline[x].B = (RGBline[x].red + RGBline[x].green)/2 - RGBline[x].blue;
	}
}

void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline);
void ConvertLineLABtoRGB(long width, LABPixel *LABline, RGBPixel *RGBline)
{	long	x;
	for (x = 0; x < width; x++)
	{	// convert from spherical LAB values to RGB
		RGBline[x].red =   LABline[x].L + LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].green = LABline[x].L - LABline[x].A/2 + LABline[x].B/3;
		RGBline[x].blue =  LABline[x].L - 2 * LABline[x].B/3;
	}
}
#define RAD 2 // neighborhood size = 5x5 (central pixel plus radius of 2 on all sides)
#define size 2*RAD+1

///////////////////////////////////
//
//	Main entry for user code
//
///////////////////////////////////
ErrType	MainUserEntry()	
{
	ErrType result = noErr;									//accumulate any errors that happened
	long	height, width;									//pixel dimensions of image
	long	x, y;											//loop counters to traverse image
	long	i, j, h, k, a, b;								//loop counters for neighborhoods
	float	minvar, meanL;									//test values for max likelihood
	float	sum1, sum2, mean, variance, count, temp;		//used to calculate variance
	float	pixel[size][size];								//pixel values
	RGBLinePtr RGBline[size];								//memory buffer - multiple lines
	LABLinePtr LABline[size];
	RGBPixel *ResultRGB;									//result line
	LABPixel *ResultLAB;

	GetOriginalDimensions(&width, &height);					//find out how big the image is

	for (i = 0; i < size; i++)								//initialize pointers
	{	RGBline[i] = NULL;
		LABline[i] = NULL;
	}
	ResultRGB = NULL;
	ResultLAB = NULL;
	
	for (i = 0; i < size; i++)
	{	RGBline[i] = CreateAPointer(width, sizeof(RGBPixel));
		LABline[i] = CreateAPointer(width, sizeof(LABPixel));	//Allocate memory for multiple rows
	}
	ResultRGB = CreateAPointer(width, sizeof(RGBPixel));	//and results
	ResultLAB = CreateAPointer(width, sizeof(LABPixel));

	for (y = 0; y < height; y++) // first pass - get min & max
	{
		if (TestAbortSetProgressBar((float)y/(float)height))
		{	//See if user canceled and update progress bar
			result = userCanceledErr;
			goto mainexit;
		}
		for (i = 0; i < size; i++)//read multiple lines
		{
			h = y + i - RAD;//vertical offset
			if (h<0) ReadOriginalLine(0, RGBline[i]); // boundary protection at top
			else if (h>=height) ReadOriginalLine(height-1, RGBline[i]);//and bottom
			else ReadOriginalLine(h, RGBline[i]);
		}
		for (i = 0; i < size; i ++)
			ConvertLineRGBtoLAB(width, RGBline[i], LABline[i]);//work with LAB values
		
		for (x = 0; x < width; x++)//scan each line pixel by pixel
		{
			for (i = 0; i < size; i++)
				for (j = 0; j < size; j++) //access neighborhood pixels
				{
					k = x + j - RAD;//horizontal offset
					if (k<0) pixel[i][j] = LABline[i][0].L;//boundary protection at left
					else if (k>=width) pixel[i][j] = LABline[i][width-1].L;//and right
					else pixel[i][j] = LABline[i][k].L;
				}
			// for each sub-nborhood calculate mean and variance, find minimum
			minvar = (float)1e20; meanL = 0;
			for (a = 0; a <= 2; a++) //a,b = nborhood position
				for (b = 0; b <= 2; b++)
				{
					sum1 = 0; sum2 = 0; count = 0;//clear sums
					for (h = 0; h <= 2; h++)//h,k = pixel posn in nborhood
						for (k = 0; k <= 2; k++)
						{
							temp = pixel[a+h][b+k];
							sum1 += temp;
							sum2 += temp*temp;
							count += 1.0;
						}
					mean = sum1/count;
					variance = (sum2/count) - (mean * mean);
					if (variance < minvar)
					{	minvar = variance;
						meanL = mean;
					}
				}
			//note that only the luminance channel is processed. this actually works well
			//in most cases, but in principle the mean color A,B values could also used.
			ResultLAB[x].L = meanL;
			ResultLAB[x].A = LABline[1][x].A;//combine with original color values
			ResultLAB[x].B = LABline[1][x].B;
		}//for x
		ConvertLineLABtoRGB(width, ResultLAB, ResultRGB);//recalculate RGB
		WriteResultLine(y, ResultRGB);//and save
	}//for y
	
mainexit:	//clean up everything we allocated
	for (i = 0; i < size; i++)
	{	if (RGBline[i]) DisposeAPointer(RGBline[i]);
		if (LABline[i]) DisposeAPointer(LABline[i]);
	}
	if (ResultRGB) DisposeAPointer(ResultRGB);
	if (ResultLAB) DisposeAPointer(ResultLAB);

	return result;											//pass any errors back
}//MainUserEntry

