
/* Photoshop plug-in example */
/*
	Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved.

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
	distributed for a fee.
*/
/* Problem 3.1.5.1 Fourier Transform Power Spectrum */

#include "PhotoshopShell.h"

//Note: This flag selects between a logarithmic and a linear display of the
//		power in the powerspectrum.
#define LOG_POWERSPECTRUM 1	//1 for log, 0 for Linear

/////////////////////SINGLE PRECISION FFT STUFF////////////////////////////////

static long gFFT_Table_Size = 0;				//global holding size of current trig tables
float  gCosTab[32768], gSinTab[32768];			//global storage for current trig tables


//Build trig tables for FFT
void PrepareFFT(long size, float *sintab, float *costab);
void PrepareFFT(long size, float *sintab, float *costab)
{
	long	angle;
	
	// only need 180 degrees, single precision FFT
	for (angle = 0; angle < size; angle++)
	{
		float theta = (float)3.141592653589793 * ((float)angle) / size;
		sintab[angle] = (float)sin(theta);
		costab[angle] = (float)cos(theta);
	}
	gFFT_Table_Size = size;	//flag to not build table each time
}//PrepareFFT


//In-Place Cooley-Tukey single precision FFT. Radix 2, DIT.  Input size must be a power-of-two: 4,8,16,32,64,...,32768
void FFT_CT(Boolean goForward, FFTelem *lptr, long size);
void FFT_CT(Boolean goForward, FFTelem *lptr, long size)
{
	FFTelem *L = (FFTelem*)&(lptr[-1]);			// allows counting 1..size instead of 0..size-1 (FORTRAN STYLE)
	float  *costab =    &(gCosTab[0]);			// local array pointers to global trig tables
	float  *sintab =    &(gSinTab[0]);			// (has to do with how compilers access global variables .. speed)

	if (gFFT_Table_Size != size)				// build trig tables if they don't exist for this size
		PrepareFFT(size, sintab, costab);
	
	{	//DECIMATION IN TIME
		long	i, 
				j = 1;
		long	m;
		for (i = 1; i <= size; i++)
		{
			if (i < j)
			{	// swap complex[i] with complex[j]
				FFTelem temp = L[j];
				L[j] = L[i];
				L[i] = temp;
			}
			m = size >> 1;
			while (j > m)
			{
				j -= m;
				m = (m+1) >> 1;
			}//while j > m
			j += m;
		}// for i
	}
	
	{	//DIVIDE AND CONQUER recursion, but implemented as a loop, Danielson-Lanczos method
		long	levelp = 1;			//steps 1, 2, 4, 8, .. size/2
		long	log2levelp = 0;		//steps 0, 1, 2, 3, .. log2(size/2)
		while (levelp < size)
		{
			long	stride = levelp * 2;
			long	m;						//used to climb levels
			for (m = 1; m <= levelp; m++)	//climb all the levels from 1..levelp
			{
				//		complex value: W = exp(-2*pi*m / levelp)
				long	w = (m - 1) * (size >> log2levelp); //w is index into cos/sin tables to look up W
				long	i;									//primary index for butterfly
				float	wr,									//real component of W
						wi;									//imaginary component of W
				
				if (goForward)								//subtle difference between forward and inverse FFT
				{
					wr = costab[w];							//using two trig tables: costab[] and sintab[]
					wi = -sintab[w];						//for performance
				}
				else
				{
					wr = costab[w];	
					wi = sintab[w];
				}
				for (i = m; i <= size; i += stride)
				{									
					long	j = i + levelp;					//second index for butterfly
					float	tempr = wr * L[j].real - wi * L[j].imag;
					float	tempi = wr * L[j].imag + wi * L[j].real;
					L[j].real = L[i].real - tempr;			//apply the "Butterfly"
					L[j].imag = L[i].imag - tempi;
					L[i].real += tempr;
					L[i].imag += tempi;
				}//for i
			}//for m loop (climbing up to levelp)
			levelp = levelp * 2;
			log2levelp++;
		}//while levelp < size
	}
	
	// recondition the inverse function
	if (!goForward)
	{
		long	i;
		for (i = 1; i <= size; i++)
		{
			L[i].real /= size;
			L[i].imag /= size;
		}
	}
}//FFT_CT


///////////////////////////////////
//
//	Main entry for user code
//
///////////////////////////////////
ErrType	MainUserEntry()	
{
	ErrType result = userCanceledErr;						//accumulate any errors that happened
	long	height, width;									//pixel dimensions of image
	long	fsize;
	long	x, y;											//loop counters to traverse image
	RGBPixel *line = NULL;									//memory buffer
	FFTelem *fftline = NULL;
	Image_ID	fft[3] = {0, 0, 0};
	long	channel;
	float	maxpower2, minpower2;							//minimum power^2 and maximum power^2
	float	scale, offset;									//values for scaling the FFT output

	GetOriginalDimensions(&width, &height);					//find out how big the image is
	fsize = height;
	if (fsize < width) fsize = width;						//find largest dimension for square FFT

	if (fsize <= 32) fsize = 32;
	else if (fsize <= 64) fsize = 64;
	else if (fsize <= 128) fsize = 128;
	else if (fsize <= 256) fsize = 256;
	else if (fsize <= 512) fsize = 512;
	else if (fsize <= 1024) fsize = 1024;
	else if (fsize <= 2048) fsize = 2048;
	else if (fsize <= 4096) fsize = 4096;
	else if (fsize <= 8192) fsize = 8192;
	else if (fsize <= 16384) fsize = 16384;
	else fsize = 32768;

	if (height != width)
	{
		ErrorMessage("Image must be SQUARE, at least 32x32, \nand a power-of-two in size");
		goto mainexit;
	}
	if (height != fsize)
	{
		ErrorMessage("Image must be SQUARE, at least 32x32, \nand a POWER-OF-TWO IN SIZE");
		goto mainexit;
	}

	line = CreateAPointer(width, sizeof(RGBPixel));		//Allocate memory for one row from either image
	fftline = CreateAPointer(fsize, sizeof(FFTelem));

	//allocate some images
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Red FFT", &(fft[0])) != noErr)
		goto mainexit;
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Green FFT", &(fft[1])) != noErr)
		goto mainexit;
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Blue FFT", &(fft[2])) != noErr)
		goto mainexit;

	//1 - separate RGB image into 3 FFT images
	//2 - perform a forwardFT on each of the 3
	//3 - find min and max power within those FFTs
	//4 - autoscale so that max power-> white, min power ->black
	//	  and produce an output image with each channel's powerspectrum in place
	{
		//clear fftline
		for (x = 0; x < fsize; x++)
			fftline[x].real = fftline[x].imag = 0.0;

		//stuff image data into appropriate FFT channels
		for (y = 0; y < height; y++)
		{
			if (TestAbortSetProgressBar(((float)y / (float)(height)) * (float)0.33333))
				goto mainexit;

			ReadOriginalLine(y, line);

			//stuff red
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].red;
			WriteTempImageLine(fft[0], y, fftline);

			//stuff green
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].green;
			WriteTempImageLine(fft[1], y, fftline);

			//stuff blue
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].blue;
			WriteTempImageLine(fft[2], y, fftline);
		}

		//process FFTs independently
		for (channel = 0; channel < 3; channel++)
		{
			//FORWARD transform
			for (y = 0; y < fsize; y++)	//horizontal
			{
				ReadTempImageLine(fft[channel], y, fftline);
				FFT_CT(TRUE, fftline, fsize);
				WriteTempImageLine(fft[channel], y, fftline);
			}
			for (x = 0; x < fsize; x++) //vertical
			{
				ReadTempImageColumn(fft[channel], x, fftline);
				FFT_CT(TRUE, fftline, fsize);
				WriteTempImageColumn(fft[channel], x, fftline);
			}
		}


		//Get min and max limits for power in the FFT images
		maxpower2 = -1e9; //these are squared terms
		minpower2 = 1e9;
		for (y = 0; y < fsize; y++)
		{
			if (TestAbortSetProgressBar(((float)y / (float)(height)) * (float)0.33333 + (float)0.33333))
				goto mainexit;

			//fetch each channel
			for (channel = 0; channel < 3; channel++)
			{
				ReadTempImageLine(fft[0], y, fftline);
				for (x = 0; x < fsize; x++)
				{	//get largest and smallest power values
					float r = fftline[x].real;
					float i = fftline[x].imag;
					float power2 = r * r + i * i;
					if (maxpower2 < power2)
						maxpower2 = power2;
					if (minpower2 > power2)
						minpower2 = power2;
				}
			}
		}

		//build a power spectrum in color using the 3 channels
		if (minpower2 == 0) minpower2 = (float)0.01;
#if LOG_POWERSPECTRUM									//select between log and linear autoscale methods
		offset = (float)log(minpower2);
		scale = (float)255.0 / (float)(log(maxpower2) - log(minpower2)); 
#else
		offset = (float)sqrt(minpower2);
		scale = (float)255.0 / (float)(sqrt(maxpower2) - sqrt(minpower2)); 
#endif
		for (y = 0; y < fsize; y++)
		{
			long	yblockswap = y + fsize / 2;			//vertical part of the blockswap
			if (yblockswap >= fsize) yblockswap -= fsize;

			if (TestAbortSetProgressBar(((float)y / (float)(height)) * (float)0.33333 + (float)0.66666))
				goto mainexit;
			ReadTempImageLine(fft[0], yblockswap, fftline);		//red
			for (x = 0; x < fsize / 2; x++)				//horizontal part of the blockswap
			{
				FFTelem swap = fftline[x];
				fftline[x] = fftline[x + fsize / 2];
				fftline[x + fsize / 2] = swap;
			}
			for (x = 0; x < width; x++)
			{
				float r = fftline[x].real;
				float i = fftline[x].imag;
				float power2 = r * r + i * i;
				if (power2 < (float)0.01) power2 = (float)0.01;	//protect against 0;
#if LOG_POWERSPECTRUM
				line[x].red = ((float)log(power2) - offset) * scale;
#else
				line[x].red = ((float)sqrt(power2) - offset) * scale;
#endif
			}

			ReadTempImageLine(fft[1], yblockswap, fftline);		//green
			for (x = 0; x < fsize / 2; x++)				//horizontal part of the blockswap
			{
				FFTelem swap = fftline[x];
				fftline[x] = fftline[x + fsize / 2];
				fftline[x + fsize / 2] = swap;
			}
			for (x = 0; x < width; x++)
			{
				float r = fftline[x].real;
				float i = fftline[x].imag;
				float power2 = r * r + i * i;
				if (power2 < (float)0.01) power2 = (float)0.01;	//protect against 0;
#if LOG_POWERSPECTRUM
				line[x].green = ((float)log(power2) - offset) * scale;
#else
				line[x].green = ((float)sqrt(power2) - offset) * scale;
#endif
			}

			ReadTempImageLine(fft[2], yblockswap, fftline);		//blue
			for (x = 0; x < fsize / 2; x++)				//horizontal part of the blockswap
			{
				FFTelem swap = fftline[x];
				fftline[x] = fftline[x + fsize / 2];
				fftline[x + fsize / 2] = swap;
			}
			for (x = 0; x < width; x++)
			{
				float r = fftline[x].real;
				float i = fftline[x].imag;
				float power2 = r * r + i * i;
				if (power2 < (float)0.01) power2 = (float)0.01;	//protect against 0;
#if LOG_POWERSPECTRUM
				line[x].blue = ((float)log(power2) - offset) * scale;
#else
				line[x].blue = ((float)sqrt(power2) - offset) * scale;
#endif
			}
			WriteResultLine(y, line);
		}
	}

	result = noErr;

mainexit:													//clean up everything we allocated
	if (line)
		DisposeAPointer(line);
	if (fftline)
		DisposeAPointer(fftline);
	//my fft images are automatically tossed for me
	return result;											//pass any errors back
}//MainUserEntry

