
/* Photoshop plug-in example */
/*
	Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved.

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
	distributed for a fee.
*/
/* Problem 3.1.5.6 Apply Mask to Fourier Transform */

#include "PhotoshopShell.h"

/////////////////////SINGLE PRECISION FFT STUFF////////////////////////////////

static long gFFT_Table_Size = 0;				//global holding size of current trig tables
float  gCosTab[32768], gSinTab[32768];			//global storage for current trig tables

//Build trig tables for FFT
void PrepareFFT(long size, float *sintab, float *costab);
void PrepareFFT(long size, float *sintab, float *costab)
{
	long	angle;
	
	// only need 180 degrees, single precision FFT
	for (angle = 0; angle < size; angle++)
	{
		float theta = (float)3.141592653589793 * ((float)angle) / size;
		sintab[angle] = (float)sin(theta);
		costab[angle] = (float)cos(theta);
	}
	gFFT_Table_Size = size;	//flag to not build table each time
}//PrepareFFT


//In-Place Cooley-Tukey single precision FFT. Radix 2, DIT.  Input size must be a power-of-two: 4,8,16,32,64,...,32768
void FFT_CT(Boolean goForward, FFTelem *lptr, long size);
void FFT_CT(Boolean goForward, FFTelem *lptr, long size)
{
	FFTelem *L = (FFTelem*)&(lptr[-1]);			// allows counting 1..size instead of 0..size-1 (FORTRAN STYLE)
	float  *costab =    &(gCosTab[0]);			// local array pointers to global trig tables
	float  *sintab =    &(gSinTab[0]);			// (has to do with how compilers access global variables .. speed)

	if (gFFT_Table_Size != size)				// build trig tables if they don't exist for this size
		PrepareFFT(size, sintab, costab);
	
	{	//DECIMATION IN TIME
		long	i, 
				j = 1;
		long	m;
		for (i = 1; i <= size; i++)
		{
			if (i < j)
			{	// swap complex[i] with complex[j]
				FFTelem temp = L[j];
				L[j] = L[i];
				L[i] = temp;
			}
			m = size >> 1;
			while (j > m)
			{
				j -= m;
				m = (m+1) >> 1;
			}//while j > m
			j += m;
		}// for i
	}
	
	{	//DIVIDE AND CONQUER recursion, but implemented as a loop, Danielson-Lanczos method
		long	levelp = 1;			//steps 1, 2, 4, 8, .. size/2
		long	log2levelp = 0;		//steps 0, 1, 2, 3, .. log2(size/2)
		while (levelp < size)
		{
			long	stride = levelp * 2;
			long	m;						//used to climb levels
			for (m = 1; m <= levelp; m++)	//climb all the levels from 1..levelp
			{
				//		complex value: W = exp(-2*pi*m / levelp)
				long	w = (m - 1) * (size >> log2levelp); //w is index into cos/sin tables to look up W
				long	i;									//primary index for butterfly
				float	wr,									//real component of W
						wi;									//imaginary component of W
				
				if (goForward)								//subtle difference between forward and inverse FFT
				{
					wr = costab[w];							//using two trig tables: costab[] and sintab[]
					wi = -sintab[w];						//for performance
				}
				else
				{
					wr = costab[w];	
					wi = sintab[w];
				}
				for (i = m; i <= size; i += stride)
				{									
					long	j = i + levelp;					//second index for butterfly
					float	tempr = wr * L[j].real - wi * L[j].imag;
					float	tempi = wr * L[j].imag + wi * L[j].real;
					L[j].real = L[i].real - tempr;			//apply the "Butterfly"
					L[j].imag = L[i].imag - tempi;
					L[i].real += tempr;
					L[i].imag += tempi;
				}//for i
			}//for m loop (climbing up to levelp)
			levelp = levelp * 2;
			log2levelp++;
		}//while levelp < size
	}
	
	// recondition the inverse function
	if (!goForward)
	{
		long	i;
		for (i = 1; i <= size; i++)
		{
			L[i].real /= size;
			L[i].imag /= size;
		}
	}
}//FFT_CT


///////////////////////////////////
//
//	Main entry for user code
//
///////////////////////////////////
ErrType	MainUserEntry()	
{
	ErrType result = userCanceledErr;						//accumulate any errors that happened
	long	height, width;									//pixel dimensions of image
	long	fsize;											//next size power-of-two that fits whole image
	long	x, y;											//loop counters to traverse image
	RGBPixel *line = NULL;									//memory buffer for image data
	FFTelem *fftline = NULL;								//memory buffer for fft data
	Image_ID	fft[3] = {0, 0, 0};							//three images -- one for each FFT channel
	long	channel;										//loop counter
	long	height2, width2;								//dimensions of reference image
	RGBPixel offset = {0.0, 0.0, 0.0};						//offset used when computing output image

	GetOriginalDimensions(&width, &height);					//find out how big the image is
	GetRefDimensions(&width2, &height2);					//and the reference image (holds the fourier mask)

	fsize = height;
	if (fsize < width)		 fsize = width;					//find largest dimension for square FFT -> fsize
	if (fsize < width2)		 fsize = width2;
	if (fsize < height2)	 fsize = height2;				//from both images

	if (fsize <= 32)		 fsize = 32;					//compute next power-of-two that will hold this image
	else if (fsize <= 64)    fsize = 64;
	else if (fsize <= 128)   fsize = 128;
	else if (fsize <= 256)   fsize = 256;
	else if (fsize <= 512)   fsize = 512;
	else if (fsize <= 1024)  fsize = 1024;
	else if (fsize <= 2048)  fsize = 2048;
	else if (fsize <= 4096)  fsize = 4096;
	else if (fsize <= 8192)  fsize = 8192;
	else if (fsize <= 16384) fsize = 16384;					//this code is limited to 32K x 32K,
	else fsize = 32768;										//but you'll run out of memory before then :)

	if (height != width)
	{
		ErrorMessage("Image must be SQUARE, at least 32x32, a power-of-two in size, and the same size as the reference image");
		goto mainexit;	//bails with userCanceledErr
	}
	if (height2 != width2)
	{
		ErrorMessage("Image must be SQUARE, at least 32x32, a power-of-two in size, and the SAME SIZE as the reference image");
		goto mainexit;	//bails with userCanceledErr
	}
	if (height != fsize)
	{
		ErrorMessage("Image must be square, at least 32x32, a POWER-OF-TWO in size, and the same size as the reference image");
		goto mainexit;	//bails with userCanceledErr
	}

	line = CreateAPointer(fsize, sizeof(RGBPixel));		//Allocate memory for one row from either image
	fftline = CreateAPointer(fsize, sizeof(FFTelem));	//And memory for one row/column from an FFT image

	//allocate three channels of FFTs so we can do each component separately
	result = MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Red FFT", &(fft[0]));
	if (result != noErr) goto mainexit;
	result = MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Green FFT", &(fft[1]));
	if (result != noErr) goto mainexit;
	result = MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Blue FFT", &(fft[2]));
	if (result != noErr) goto mainexit;

	//1. split image into three separate channels
	//2. apply forwardFT to each channel
	//3. apply mask to each transformed channel
	//4. apply inverseFT to each channel
	//5. recombine channels to form output image
	{
		//clear fftline
		for (x = 0; x < fsize; x++)
			fftline[x].real = fftline[x].imag = 0.0;

		//stuff image data into appropriate FFT channels
		for (y = 0; y < height; y++)
		{
			if (TestAbortSetProgressBar((float)y / (float)(height)))
				goto mainexit;	//bails with userCanceledErr

			ReadOriginalLine(y, line);

			//stuff red
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].red;
			WriteTempImageLine(fft[0], y, fftline);

			//stuff green
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].green;
			WriteTempImageLine(fft[1], y, fftline);

			//stuff blue
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].blue;
			WriteTempImageLine(fft[2], y, fftline);
		}

		//process FFTs independently
		for (channel = 0; channel < 3; channel++)
		{
			if (TestAbortSetProgressBar((float)channel / (float)3))
				goto mainexit;	//bails with userCanceledErr

			//FORWARD transform
			for (y = 0; y < fsize; y++)	//horizontal
			{
				ReadTempImageLine(fft[channel], y, fftline);
				FFT_CT(TRUE, fftline, fsize);
				WriteTempImageLine(fft[channel], y, fftline);
			}
			for (x = 0; x < fsize; x++) //vertical
			{
				ReadTempImageColumn(fft[channel], x, fftline);
				FFT_CT(TRUE, fftline, fsize);
				WriteTempImageColumn(fft[channel], x, fftline);
			}
		}
		//NOTE: DC is at 0,0 in the fft images.  If DC in the mask == 0 then we need to add midpoint offset to the whole image
		ReadRefImageLine(fsize / 2, line);
		if (line[fsize / 2].red < 0.5)		offset.red = 127.5;
		if (line[fsize / 2].green < 0.5)	offset.green = 127.5;
		if (line[fsize / 2].blue < 0.5)		offset.blue = 127.5;

		//apply reference image mask, in place, on FFT images (do blockswap on mask)
		for (y = 0; y < height2; y++)
		{
			long yblockswap = y + height2 / 2;
			if (yblockswap >= height2) yblockswap -= height2;		//vertical blockswap

			if (TestAbortSetProgressBar((float)y / (float)height2))
				goto mainexit;	//bails with userCanceledErr

			ReadRefImageLine(yblockswap, line);
			for (x = 0; x < width2 / 2; x++)						//horizontal blockswap
			{
				RGBPixel swap = line[x];
				line[x] = line[x + width2 / 2];
				line[x + width2 / 2] = swap;
			}

			//apply mask channel by channel
			ReadTempImageLine(fft[0], y, fftline); //red
			for (x = 0; x < width2; x++)
			{
				fftline[x].imag *= line[x].red / (float)255.0;
				fftline[x].real *= line[x].red / (float)255.0;
			}
			WriteTempImageLine(fft[0], y, fftline);

			ReadTempImageLine(fft[1], y, fftline); //green
			for (x = 0; x < width2; x++)
			{
				fftline[x].imag *= line[x].green / (float)255.0;
				fftline[x].real *= line[x].green / (float)255.0;
			}
			WriteTempImageLine(fft[1], y, fftline);

			ReadTempImageLine(fft[2], y, fftline); //blue
			for (x = 0; x < width2; x++)
			{
				fftline[x].imag *= line[x].blue / (float)255.0;
				fftline[x].real *= line[x].blue / (float)255.0;
			}
			WriteTempImageLine(fft[2], y, fftline);
		}

		//INVERSE transform on each channel
		for (channel = 0; channel < 3; channel++)
		{
			if (TestAbortSetProgressBar((float)channel / (float)3)) 
				goto mainexit;	//bails with userCanceledErr

			for (y = 0; y < fsize; y++)	//horizontal
			{
				ReadTempImageLine(fft[channel], y, fftline);
				FFT_CT(FALSE, fftline, fsize);
				WriteTempImageLine(fft[channel], y, fftline);
			}
			for (x = 0; x < fsize; x++) //vertical
			{
				ReadTempImageColumn(fft[channel], x, fftline);
				FFT_CT(FALSE, fftline, fsize);
				WriteTempImageColumn(fft[channel], x, fftline);
			}
		}

		//copy out all 3 channels back into main image
		for (y = 0; y < height; y++)
		{
			if (TestAbortSetProgressBar((float)y / (float)(height))) 
				goto mainexit;	//bails with userCanceledErr

			//fetch red
			ReadTempImageLine(fft[0], y, fftline);
			for (x = 0; x < width; x++)
				line[x].red = fftline[x].real + offset.red;

			//fetch green
			ReadTempImageLine(fft[1], y, fftline);
			for (x = 0; x < width; x++)
				line[x].green = fftline[x].real + offset.green;

			//fetch blue
			ReadTempImageLine(fft[2], y, fftline);
			for (x = 0; x < width; x++)
				line[x].blue = fftline[x].real + offset.blue;

			WriteResultLine(y, line);
		}
	}
	result = noErr;	//pass out success

mainexit:													//clean up everything we allocated
	if (line)
		DisposeAPointer(line);
	if (fftline)
		DisposeAPointer(fftline);
	//my fft images are automatically tossed for me
	//and the Reference image is automatically closed
	return result;											//pass any errors back
}//MainUserEntry

