
/* Photoshop plug-in example */
/*
	Copyright (c) 1996-2007, J. Christian Russ and John C. Russ
	All rights reserved.

	This code may be used by purchasers of the book for educational and personal research
	purposes but not for any commercial purpose.  The Copyright notices must be preserved
	in all copies. Plug-ins utilizing or derived from this code may not be sold or 
	distributed for a fee.
*/
/* Problem 3.3.3.3 Cross Correlate Two Images */

#include "PhotoshopShell.h"

/////////////////////SINGLE PRECISION FFT STUFF////////////////////////////////

static long gFFT_Table_Size = 0;				//global holding size of current trig tables
float  gCosTab[32768], gSinTab[32768];			//global storage for current trig tables


//Build trig tables for FFT
void PrepareFFT(long size, float *sintab, float *costab);
void PrepareFFT(long size, float *sintab, float *costab)
{
	long	angle;
	
	// only need 180 degrees, single precision FFT
	for (angle = 0; angle < size; angle++)
	{
		float theta = (float)3.141592653589793 * ((float)angle) / size;
		sintab[angle] = (float)sin(theta);
		costab[angle] = (float)cos(theta);
	}
	gFFT_Table_Size = size;	//flag to not build table each time
}//PrepareFFT


//In-Place Cooley-Tukey single precision FFT. Radix 2, DIT.  Input size must be a power-of-two: 4,8,16,32,64,...,32768
void FFT_CT(Boolean goForward, FFTelem *lptr, long size);
void FFT_CT(Boolean goForward, FFTelem *lptr, long size)
{
	FFTelem *L = (FFTelem*)&(lptr[-1]);			// allows counting 1..size instead of 0..size-1 (FORTRAN STYLE)
	float  *costab =    &(gCosTab[0]);			// local array pointers to global trig tables
	float  *sintab =    &(gSinTab[0]);			// (has to do with how compilers access global variables .. speed)

	if (gFFT_Table_Size != size)				// build trig tables if they don't exist for this size
		PrepareFFT(size, sintab, costab);
	
	{	//DECIMATION IN TIME
		long	i, 
				j = 1;
		long	m;
		for (i = 1; i <= size; i++)
		{
			if (i < j)
			{	// swap complex[i] with complex[j]
				FFTelem temp = L[j];
				L[j] = L[i];
				L[i] = temp;
			}
			m = size >> 1;
			while (j > m)
			{
				j -= m;
				m = (m+1) >> 1;
			}//while j > m
			j += m;
		}// for i
	}
	
	{	//DIVIDE AND CONQUER recursion, but implemented as a loop, Danielson-Lanczos method
		long	levelp = 1;			//steps 1, 2, 4, 8, .. size/2
		long	log2levelp = 0;		//steps 0, 1, 2, 3, .. log2(size/2)
		while (levelp < size)
		{
			long	stride = levelp * 2;
			long	m;						//used to climb levels
			for (m = 1; m <= levelp; m++)	//climb all the levels from 1..levelp
			{
				//		complex value: W = exp(-2*pi*m / levelp)
				long	w = (m - 1) * (size >> log2levelp); //w is index into cos/sin tables to look up W
				long	i;									//primary index for butterfly
				float	wr,									//real component of W
						wi;									//imaginary component of W
				
				if (goForward)								//subtle difference between forward and inverse FFT
				{
					wr = costab[w];							//using two trig tables: costab[] and sintab[]
					wi = -sintab[w];						//for performance
				}
				else
				{
					wr = costab[w];	
					wi = sintab[w];
				}
				for (i = m; i <= size; i += stride)
				{									
					long	j = i + levelp;					//second index for butterfly
					float	tempr = wr * L[j].real - wi * L[j].imag;
					float	tempi = wr * L[j].imag + wi * L[j].real;
					L[j].real = L[i].real - tempr;			//apply the "Butterfly"
					L[j].imag = L[i].imag - tempi;
					L[i].real += tempr;
					L[i].imag += tempi;
				}//for i
			}//for m loop (climbing up to levelp)
			levelp = levelp * 2;
			log2levelp++;
		}//while levelp < size
	}
	
	// recondition the inverse function
	if (!goForward)
	{
		long	i;
		for (i = 1; i <= size; i++)
		{
			L[i].real /= size;
			L[i].imag /= size;
		}
	}
}//FFT_CT



///////////////////////////////////
//
//	Main entry for user code
//
///////////////////////////////////
ErrType	MainUserEntry()	
{
	ErrType result = userCanceledErr;						//accumulate any errors that happened
	long	height, width;									//pixel dimensions of image
	long	fsize;
	long	x, y;											//loop counters to traverse image
	RGBPixel *line = NULL;									//memory buffer
	FFTelem *fftline = NULL;
	FFTelem *fftline2 = NULL;
	Image_ID	fft[6] = {0, 0, 0, 0, 0, 0};				//0, 1, 2 are primary image, 3, 4, 5 are reference image
	long	channel;
	long	height2, width2;								//dimensions of reference image
	float	minvalue, maxvalue;								//limits after the convolution
	float	scale, offset;									//for autoscaling the output

	GetOriginalDimensions(&width, &height);					//find out how big the image is
	GetRefDimensions(&width2, &height2);					//and the reference image (holds the fourier mask)
	fsize = height;
	if (fsize < width) fsize = width;						//find largest dimension for square FFT
	if (fsize < width2) fsize = width2;
	if (fsize < height2) fsize = height2;					//from both images

	if (fsize <= 32) fsize = 32;
	else if (fsize <= 64) fsize = 64;
	else if (fsize <= 128) fsize = 128;
	else if (fsize <= 256) fsize = 256;
	else if (fsize <= 512) fsize = 512;
	else if (fsize <= 1024) fsize = 1024;
	else if (fsize <= 2048) fsize = 2048;
	else if (fsize <= 4096) fsize = 4096;
	else if (fsize <= 8192) fsize = 8192;
	else if (fsize <= 16384) fsize = 16384;
	else fsize = 32768;

	if (height != width)
	{
		ErrorMessage("Image must be SQUARE, at least 32x32, a power-of-two in size, and the same size as the reference image");
		goto mainexit;
	}
	if (height2 != width2)
	{
		ErrorMessage("Image must be SQUARE, at least 32x32, a power-of-two in size, and the SAME SIZE as the reference image");
		goto mainexit;
	}
	if (height != fsize)
	{
		ErrorMessage("Image must be square, at least 32x32, a POWER-OF-TWO in size, and the same size as the reference image");
		goto mainexit;
	}

	line = CreateAPointer(fsize, sizeof(RGBPixel));		//Allocate memory for one row from either image
	fftline = CreateAPointer(fsize, sizeof(FFTelem));
	fftline2 = CreateAPointer(fsize, sizeof(FFTelem));

	//allocate some images
	//we'll need to have FFT versions of all 3 channels (RGB) of the primary image AND the reference image
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Red FFT", &(fft[0])) != noErr)
		goto mainexit;
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Green FFT", &(fft[1])) != noErr)
		goto mainexit;
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "Blue FFT", &(fft[2])) != noErr)
		goto mainexit;
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "ref Red FFT", &(fft[3])) != noErr)
		goto mainexit;
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "ref Green FFT", &(fft[4])) != noErr)
		goto mainexit;
	if 	(MakeTemporaryImage(fsize, fsize, sizeof(FFTelem), "ref Blue FFT", &(fft[5])) != noErr)
		goto mainexit;

	//1 - split primary image into 3 channels for FFTs
	//2 - split secondary image into 3 channels for FFTs (so the array goes to 6)
	//3 - do all 6 ForwardFTs 
	//4 - perform cross-correlation, channel by channel R, G, B
	//		NOTE: two ways to perform cross-correlation:
	//			  A) multiply (convolve) with the complex conjugate (a - ib) instead of (a + ib)
	//			  B) rotate second image by 180 degrees and do normal convolution
	//			  !! we'll do method A, below
	//5 - perform InverseFT on 3 convolved channels
	//6 - find min/max values in the convolution result so we can have an output image
	//7 - produced combined RGB output image (autoscaled)
	{
		//clear fftline
		for (x = 0; x < fsize; x++)
			fftline[x].real = fftline[x].imag = 0.0;

		//stuff primary image data into appropriate FFT channels
		for (y = 0; y < height; y++)
		{
			if (TestAbortSetProgressBar((float)y / (float)(height) / (float)10.0))
				goto mainexit;

			ReadOriginalLine(y, line);

			//stuff red
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].red;
			WriteTempImageLine(fft[0], y, fftline);

			//stuff green
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].green;
			WriteTempImageLine(fft[1], y, fftline);

			//stuff blue
			for (x = 0; x < width; x++)
				fftline[x].real = line[x].blue;
			WriteTempImageLine(fft[2], y, fftline);
		}

		//clear fftline again
		for (x = 0; x < fsize; x++)
			fftline[x].real = fftline[x].imag = 0.0;

		//stuff reference image data into appropriate FFT channels
		for (y = 0; y < height; y++)
		{
			ReadRefImageLine(y, line);

			//stuff red
			for (x = 0; x < width2; x++)
				fftline[x].real = line[x].red;
			WriteTempImageLine(fft[3], y, fftline);

			//stuff green
			for (x = 0; x < width2; x++)
				fftline[x].real = line[x].green;
			WriteTempImageLine(fft[4], y, fftline);

			//stuff blue
			for (x = 0; x < width2; x++)
				fftline[x].real = line[x].blue;
			WriteTempImageLine(fft[5], y, fftline);
		}

		//process FFTs independently
		for (channel = 0; channel < 6; channel++)
		{
			//FORWARD transform
			for (y = 0; y < fsize; y++)	//horizontal
			{
				ReadTempImageLine(fft[channel], y, fftline);
				FFT_CT(TRUE, fftline, fsize);
				WriteTempImageLine(fft[channel], y, fftline);
			}
			for (x = 0; x < fsize; x++) //vertical
			{
				ReadTempImageColumn(fft[channel], x, fftline);
				FFT_CT(TRUE, fftline, fsize);
				WriteTempImageColumn(fft[channel], x, fftline);
			}
		}// for each channel

		//Fourier convolution - no blockswap was necessary
		for (channel = 0; channel < 3; channel++)
			for (y = 0; y < fsize; y++)
			{
				ReadTempImageLine(fft[channel], y, fftline);
				ReadTempImageLine(fft[channel+3], y, fftline2);
				for (x = 0; x < fsize; x++)	//convolution is complex multiplication in Fourier space
				{
					float	fftl2x_real_cc = fftline2[x].real;		//method A, use complex conjugate
					float	fftl2x_imag_cc = -fftline2[x].imag;		//to perform cross-correlation instead of convolution
					//(a + bi) * (c + di) 
					// = ac - bd + i(bc + ad)
					float	tempr = fftline[x].real * fftl2x_real_cc - fftline[x].imag * fftl2x_imag_cc;
					float	tempi = fftline[x].imag * fftl2x_real_cc + fftline[x].real * fftl2x_imag_cc;
					fftline[x].real = tempr;
					fftline[x].imag = tempi;
				}
				WriteTempImageLine(fft[channel], y, fftline);
			}

		minvalue = (float)1e38;				//get limits on the output data so we can autoscale the result
		maxvalue = (float)-1e38;
		//INVERSE transform on primary image, only
		for (channel = 0; channel < 3; channel++)
		{
			for (y = 0; y < fsize; y++)	//horizontal
			{
				ReadTempImageLine(fft[channel], y, fftline);
				FFT_CT(FALSE, fftline, fsize);
				WriteTempImageLine(fft[channel], y, fftline);
			}
			for (x = 0; x < fsize; x++) //vertical
			{
				ReadTempImageColumn(fft[channel], x, fftline);
				FFT_CT(FALSE, fftline, fsize);
				for (y = 0; y < width; y++)
				{
					if (minvalue > fftline[y].real)
						minvalue = fftline[y].real;
					if (maxvalue < fftline[y].real)
						maxvalue = fftline[y].real;
				}
				WriteTempImageColumn(fft[channel], x, fftline);
			}
		}

		//compute scale and offset for autoscaling the result
		scale = (float)255.0 / (maxvalue - minvalue);
		offset = minvalue;
		//copy out all 3 channels back into main image, will have to do a blockswap on output
		for (y = 0; y < height; y++)
		{
			long	yblockswap = y + fsize / 2;				//vertical blockswap
			if (yblockswap >= fsize)	yblockswap -= fsize;
			if (TestAbortSetProgressBar((float)y / (float)(height) / (float)10.0 + (float)0.9)) 
				goto mainexit;

			//fetch red
			ReadTempImageLine(fft[0], yblockswap, fftline);
			for (x = 0; x < width; x++)
				line[x].red = scale * (fftline[x].real - offset);

			//fetch green
			ReadTempImageLine(fft[1], yblockswap, fftline);
			for (x = 0; x < width; x++)
				line[x].green = scale * (fftline[x].real - offset);

			//fetch blue
			ReadTempImageLine(fft[2], yblockswap, fftline);
			for (x = 0; x < width; x++)
				line[x].blue = scale * (fftline[x].real - offset);

			for (x = 0; x < fsize / 2; x++)						//horizontal blockswap
			{
				RGBPixel	swap = line[x];
				line[x] = line[x + fsize / 2];
				line[x + fsize / 2] = swap;
			}

			WriteResultLine(y, line);
		}
	}

	result = noErr;

mainexit:													//clean up everything we allocated
	if (line)
		DisposeAPointer(line);
	if (fftline)
		DisposeAPointer(fftline);
	if (fftline2)
		DisposeAPointer(fftline2);
	//my fft images are automatically tossed for me
	return result;											//pass any errors back
}//MainUserEntry

